# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.fields import ScalarField, VectorField
from kgsim.simulation import GenericSimulation
from kgsim.athena.athena_read import athdf

from kbasic.parsing import Folder
from glob import glob 
import numpy as np
import builtins

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
def get_mhd_config(mhd_run_dir, config_name):
    """Get MHD run information from the configuration file

    Arguments:
        mhd_run_dir (string): MHD run directory
        config_name (string): MHD simulation configuration file name
    """
    with open(mhd_run_dir + "/" + config_name) as f:
        contents = f.readlines()
    f.close()
    mhd_config = {}
    for line in contents:
        if "<" in line and ">" in line and "<" == line[0]:
            block_name = line[1:line.find(">")]
            mhd_config[block_name] = {}
        else:
            if line[0] != "#" and "=" in line:
                line_splits = line.split("=")
                tail = line_splits[1].split("\n")
                data = tail[0].split("#")
                ltmp = line_splits[0].strip()
                try:
                    mhd_config[block_name][ltmp] = float(data[0])
                except ValueError:
                    mhd_config[block_name][ltmp] = data[0].strip()
    return mhd_config

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class AthenaParameter(ScalarField):
    def __init__(
        self, 
        param_key: str,
        parent = None,
        caching: bool = False,
        verbose: bool = False,
        name: str = None, 
        latex: str = None
    ):
        super().__init__( 
            None, 
            parent=parent, 
            caching=caching, 
            verbose=verbose, 
            name=name, 
            latex=latex
        )
        self.param_key = param_key
        self.path = parent.path
        self.single = False 
        self.file_names = parent.file_names

    def reader(self, fname: str, item: int):
        return athdf(fname, quantities=[self.param_key])[self.param_key][0]

class Athena(GenericSimulation):
    def __init__(
            self, 
            path, 
            template = None, 
            caching = False, 
            verbose = True
    ):
        super().__init__(path, template, caching, verbose)
        self.input = get_mhd_config(self.path, "athinput.reconnection")
        self.parse_output()

    def parse_output(self):
        self.file_names = sorted(glob(self.path+"/*.athdf"))
        time, x, y, dxm, dym = [],[],[],[],[]
        for file in self.file_names:
            fdata = athdf(file, quantities=[])
            time.append(fdata['Time'])
            x.append(fdata["x1f"])
            y.append(fdata["x2f"])
            dxm.append(fdata["x1f"][1] - fdata["x1f"][0])
            dym.append(fdata["x2f"][1] - fdata["x2f"][0])
        self.time = np.array(time)
        self.dx = dxm[0]
        self.dy = dym[0]
        kwargs = {'caching':self.caching, 'verbose':self.verbose, 'parent':self}
        self.B = VectorField(
            AthenaParameter("Bcc1", name="B_x", latex="$B_x$", **kwargs),
            AthenaParameter("Bcc2", name="B_y", latex="$B_y$", **kwargs),
            AthenaParameter("Bcc3", name="B_z", latex="$B_z$", **kwargs),
            name='magnetic', latex=r"$\vec{B}$", **kwargs
        )
        self.u = VectorField(
            AthenaParameter("vel1", name="u_x", latex="$u_x$", **kwargs),
            AthenaParameter("vel2", name="u_y", latex="$u_y$", **kwargs),
            AthenaParameter("vel3", name="u_z", latex="$u_z$", **kwargs),
            name='velocity', latex=r"$\vec{u}$", **kwargs
        )
        self.density = AthenaParameter("rho", name='density', latex=r"$\rho$", **kwargs)
        self.P = AthenaParameter("press", name="pressure", latex="$P$", **kwargs)


"""
Read Athena++ output data files. got from Giulia
"""

# Python modules
import re
import struct
import sys
import warnings
from io import open  # Consistent binary I/O from Python 2 and 3

# Other Python modules
import numpy as np

check_nan_flag = False


# ========================================================================================

def check_nan(data):
    """Check input NumPy array for the presence of any NaN entries"""
    if np.isnan(data).any():
        raise FloatingPointError("NaN encountered")
    return


# ========================================================================================

def error_dat(filename, **kwargs):
    """Wrapper to np.loadtxt() for applying optional checks used in regression tests"""
    data = np.loadtxt(filename,
                      dtype=np.float64,
                      ndmin=2,  # prevent NumPy from squeezing singleton dimensions
                      **kwargs)
    if check_nan_flag:
        check_nan(data)
    return data


# ========================================================================================

def hst(filename, raw=False):
    """Read .hst files and return dict of 1D arrays.


    Keyword arguments:
    raw -- if True, do not prune file to remove stale data from prev runs (default False)
    """

    # Read data
    with open(filename, 'r') as data_file:
        # Find header
        header_found = False
        multiple_headers = False
        header_location = None
        line = data_file.readline()
        while len(line) > 0:
            if line == '# Athena++ history data\n':
                if header_found:
                    multiple_headers = True
                else:
                    header_found = True
                header_location = data_file.tell()
            line = data_file.readline()
        if multiple_headers:
            warnings.warn('Multiple headers found; using most recent data', AthenaWarning)
        if header_location is None:
            raise AthenaError('No header found')

        # Parse header
        data_file.seek(header_location)
        header = data_file.readline()
        data_names = re.findall(r'\[\d+\]=(\S+)', header)
        if len(data_names) == 0:
            raise AthenaError('Header could not be parsed')

        # Prepare dictionary of results
        data = {}
        for name in data_names:
            data[name] = []

        # Read data
        for line in data_file:
            for name, val in zip(data_names, line.split()):
                data[name].append(float(val))

    # Finalize data
    for key, val in data.items():
        data[key] = np.array(val)
    if not raw:
        if data_names[0] != 'time':
            raise AthenaError('Cannot remove spurious data because time column could not'
                              + ' be identified')
        branches_removed = False
        while not branches_removed:
            branches_removed = True
            for n in range(1, len(data['time'])):
                if data['time'][n] <= data['time'][n-1]:
                    branch_index = np.where(data['time'][:n] >= data['time'][n])[0][0]
                    for key, val in data.items():
                        data[key] = np.concatenate((val[:branch_index], val[n:]))
                    branches_removed = False
                    break
        if check_nan_flag:
            for key, val in data.items():
                check_nan(val)
    return data


# ========================================================================================

def tab(filename, raw=False, dimensions=None):
    """Read .tab files and return dict or array.


    Keyword arguments:
    raw -- if True, do not parse the header to figure out key names (default False)
    """

    # Check for valid number of dimensions
    if raw and not (dimensions == 1 or dimensions == 2 or dimensions == 3):
        raise AthenaError('Improper number of dimensions')
    if not raw and dimensions is not None:
        warnings.warn('Ignoring specified number of dimensions', AthenaWarning)

    # Parse header
    if not raw:
        data_dict = {}
        with open(filename, 'r') as data_file:
            line = data_file.readline()
            attributes = re.search(r'time=(\S+)\s+cycle=(\S+)\s+variables=(\S+)', line)
            line = data_file.readline()
            headings = line.split()[1:]
        data_dict['time'] = float(attributes.group(1))
        data_dict['cycle'] = int(attributes.group(2))
        data_dict['variables'] = attributes.group(3)
        if headings[0] == 'i' and headings[2] == 'j' and headings[4] == 'k':
            headings = headings[1:2] + headings[3:4] + headings[5:]
            dimensions = 3
        elif ((headings[0] == 'i' and headings[2] == 'j')
              or (headings[0] == 'i' and headings[2] == 'k')
              or (headings[0] == 'j' and headings[2] == 'k')):
            headings = headings[1:2] + headings[3:]
            dimensions = 2
        elif headings[0] == 'i' or headings[0] == 'j' or headings[0] == 'k':
            headings = headings[1:]
            dimensions = 1
        else:
            raise AthenaError('Could not parse header')

    # Go through lines
    data_array = []
    with open(filename, 'r') as data_file:
        first_line = True
        for line in data_file:
            # Skip comments
            if line.split()[0][0] == '#':
                continue

            # Extract cell indices
            vals = line.split()
            if first_line:
                i_min = i_max = int(vals[0])
                if dimensions == 2 or dimensions == 3:
                    j_min = j_max = int(vals[2])
                if dimensions == 3:
                    k_min = k_max = int(vals[4])
                num_entries = len(vals) - dimensions
                first_line = False
            else:
                i_max = max(i_max, int(vals[0]))
                if dimensions == 2 or dimensions == 3:
                    j_max = max(j_max, int(vals[2]))
                if dimensions == 3:
                    k_max = max(k_max, int(vals[4]))

            # Extract cell values
            if dimensions == 1:
                vals = vals[1:]
            if dimensions == 2:
                vals = vals[1:2] + vals[3:]
            if dimensions == 3:
                vals = vals[1:2] + vals[3:4] + vals[5:]
            data_array.append([float(val) for val in vals])

    # Reshape array based on number of dimensions
    if dimensions == 1:
        array_shape = (i_max-i_min+1, num_entries)
        array_transpose = (1, 0)
    if dimensions == 2:
        array_shape = (j_max-j_min+1, i_max-i_min+1, num_entries)
        array_transpose = (2, 0, 1)
    if dimensions == 3:
        array_shape = (k_max-k_min+1, j_max-j_min+1, i_max-i_min+1, num_entries)
        array_transpose = (3, 0, 1, 2)
    data_array = np.transpose(np.reshape(data_array, array_shape), array_transpose)

    # Finalize data
    if raw:
        if check_nan_flag:
            check_nan(data_array)
        return data_array
    else:
        for n, heading in enumerate(headings):
            if check_nan_flag:
                check_nan(data_array[n, ...])
            data_dict[heading] = data_array[n, ...]
        return data_dict


# ========================================================================================

def vtk(filename):
    """Read .vtk files and return dict of arrays of data."""

    # Read raw data
    with open(filename, 'rb') as data_file:
        raw_data = data_file.read()
    raw_data_ascii = raw_data.decode('ascii', 'replace')

    # Skip header
    current_index = 0
    current_char = raw_data_ascii[current_index]
    while current_char == '#':
        while current_char != '\n':
            current_index += 1
            current_char = raw_data_ascii[current_index]
        current_index += 1
        current_char = raw_data_ascii[current_index]

    # Function for skipping though the file
    def skip_string(expected_string):
        expected_string_len = len(expected_string)
        if (raw_data_ascii[current_index:current_index+expected_string_len]
                != expected_string):
            raise AthenaError('File not formatted as expected')
        return current_index+expected_string_len

    # Read metadata
    current_index = skip_string('BINARY\nDATASET RECTILINEAR_GRID\nDIMENSIONS ')
    end_of_line_index = current_index + 1
    while raw_data_ascii[end_of_line_index] != '\n':
        end_of_line_index += 1
    data_to_map = raw_data_ascii[current_index:end_of_line_index]
    face_dimensions = list(map(int, data_to_map.split(' ')))
    current_index = end_of_line_index + 1

    # Function for reading interface locations
    def read_faces(letter, num_faces):
        identifier_string = '{0}_COORDINATES {1} float\n'.format(letter, num_faces)
        begin_index = skip_string(identifier_string)
        format_string = '>' + 'f'*num_faces
        end_index = begin_index + 4*num_faces
        vals = np.array(struct.unpack(format_string, raw_data[begin_index:end_index]))
        return vals, end_index+1

    # Read interface locations
    x_faces, current_index = read_faces('X', face_dimensions[0])
    y_faces, current_index = read_faces('Y', face_dimensions[1])
    z_faces, current_index = read_faces('Z', face_dimensions[2])

    # Prepare to read quantities defined on grid
    cell_dimensions = np.array([max(dim-1, 1) for dim in face_dimensions])
    num_cells = cell_dimensions.prod()
    current_index = skip_string('CELL_DATA {0}\n'.format(num_cells))
    if raw_data_ascii[current_index:current_index+1] == '\n':
        current_index = skip_string('\n')  # extra newline inserted by join script
    data = {}

    # Function for reading scalar data
    def read_cell_scalars():
        begin_index = skip_string('SCALARS ')
        end_of_word_index = begin_index + 1
        while raw_data_ascii[end_of_word_index] != ' ':
            end_of_word_index += 1
        array_name = raw_data_ascii[begin_index:end_of_word_index]
        string_to_skip = 'SCALARS {0} float\nLOOKUP_TABLE default\n'.format(array_name)
        begin_index = skip_string(string_to_skip)
        format_string = '>' + 'f'*num_cells
        end_index = begin_index + 4*num_cells
        data[array_name] = struct.unpack(format_string, raw_data[begin_index:end_index])
        dimensions = tuple(cell_dimensions[::-1])
        data[array_name] = np.array(data[array_name]).reshape(dimensions)
        return end_index+1

    # Function for reading vector data
    def read_cell_vectors():
        begin_index = skip_string('VECTORS ')
        end_of_word_index = begin_index + 1
        while raw_data_ascii[end_of_word_index] != '\n':
            end_of_word_index += 1
        array_name = raw_data_ascii[begin_index:end_of_word_index]
        string_to_skip = 'VECTORS {0}\n'.format(array_name)
        array_name = array_name[:-6]  # remove ' float'
        begin_index = skip_string(string_to_skip)
        format_string = '>' + 'f'*num_cells*3
        end_index = begin_index + 4*num_cells*3
        data[array_name] = struct.unpack(format_string, raw_data[begin_index:end_index])
        dimensions = tuple(np.append(cell_dimensions[::-1], 3))
        data[array_name] = np.array(data[array_name]).reshape(dimensions)
        return end_index+1

    # Read quantities defined on grid
    while current_index < len(raw_data):
        expected_string = 'SCALARS'
        expected_string_len = len(expected_string)
        if (raw_data_ascii[current_index:current_index+expected_string_len]
                == expected_string):
            current_index = read_cell_scalars()
            continue
        expected_string = 'VECTORS'
        expected_string_len = len(expected_string)
        if (raw_data_ascii[current_index:current_index+expected_string_len]
                == expected_string):
            current_index = read_cell_vectors()
            continue
        raise AthenaError('File not formatted as expected')

    if check_nan_flag:
        check_nan(x_faces)
        check_nan(y_faces)
        check_nan(z_faces)
        for key, val in data.items():
            check_nan(val)

    return x_faces, y_faces, z_faces, data


# ========================================================================================

def athdf(filename, raw=False, data=None, quantities=None, dtype=None, level=None,
          return_levels=False, subsample=False, fast_restrict=False, x1_min=None,
          x1_max=None, x2_min=None, x2_max=None, x3_min=None, x3_max=None, vol_func=None,
          vol_params=None, face_func_1=None, face_func_2=None, face_func_3=None,
          center_func_1=None, center_func_2=None, center_func_3=None, num_ghost=0):
    """Read .athdf files and populate dict of arrays of data.


    Keyword arguments:
    raw -- if True, do not merge MeshBlocks into a single array (default False)
    """

    # Load HDF5 reader
    import h5py

    # Handle request for raw data
    if raw:
        # Open file
        with h5py.File(filename, 'r') as f:
            # Store file-level attributes
            data = {}
            for key in f.attrs:
                data[str(key)] = f.attrs[key]

            # Store location metadata
            data['Levels'] = f['Levels'][:]
            data['LogicalLocations'] = f['LogicalLocations'][:]

            # Store coordinate data
            data['x1f'] = f['x1f'][:]
            data['x2f'] = f['x2f'][:]
            data['x3f'] = f['x3f'][:]
            data['x1v'] = f['x1v'][:]
            data['x2v'] = f['x2v'][:]
            data['x3v'] = f['x3v'][:]

            # Get metadata describing file layout
            dataset_names = np.array([x.decode('ascii', 'replace')
                                      for x in f.attrs['DatasetNames'][:]])
            dataset_sizes = f.attrs['NumVariables'][:]
            variable_names = np.array([x.decode('ascii', 'replace')
                                       for x in f.attrs['VariableNames'][:]])

            # Store cell data
            for dataset_index, dataset_name in enumerate(dataset_names):
                variable_begin = sum(dataset_sizes[:dataset_index])
                variable_end = variable_begin + dataset_sizes[dataset_index]
                variable_names_local = variable_names[variable_begin:variable_end]
                for variable_index, variable_name in enumerate(variable_names_local):
                    data[variable_name] = f[dataset_name][variable_index, ...]

        if check_nan_flag:
            for key, val in data.items():
                if key in variable_names:
                    check_nan(val)
        # Return dictionary containing raw data
        return data

    # Prepare dictionary for results
    if data is None:
        data = {}
        new_data = True
    else:
        new_data = False

    # Open file
    with h5py.File(filename, 'r') as f:
        # Extract size information
        max_level = f.attrs['MaxLevel']
        if level is None:
            level = max_level
        block_size = f.attrs['MeshBlockSize']
        root_grid_size = f.attrs['RootGridSize']
        levels = f['Levels'][:]
        logical_locations = f['LogicalLocations'][:]
        if dtype is None:
            dtype = f[f.attrs['DatasetNames'][0]].dtype.newbyteorder('=')
        if num_ghost == 0 and np.array(f['x1v']).min() < f.attrs['RootGridX1'][0]:
            raise AthenaError('Ghost zones detected but "num_ghost" keyword set to zero.')
        if num_ghost > 0 and not np.all(levels == max_level):
            raise AthenaError('Cannot use ghost zones with different refinement levels')
        nx_vals = []
        for d in range(3):
            if block_size[d] == 1 and root_grid_size[d] > 1:  # sum or slice
                other_locations = [location
                                   for location in zip(levels,
                                                       logical_locations[:, (d+1) % 3],
                                                       logical_locations[:, (d+2) % 3])]
                if len(set(other_locations)) == len(other_locations):  # effective slice
                    nx_vals.append(1)
                else:  # nontrivial sum
                    num_blocks_this_dim = 0
                    for level_this_dim, loc_this_dim in zip(levels,
                                                            logical_locations[:, d]):
                        if level_this_dim <= level:
                            possible_max = (loc_this_dim+1) * 2**(level-level_this_dim)
                            num_blocks_this_dim = max(num_blocks_this_dim, possible_max)
                        else:
                            possible_max = (loc_this_dim+1) // 2**(level_this_dim-level)
                            num_blocks_this_dim = max(num_blocks_this_dim, possible_max)
                    nx_vals.append(num_blocks_this_dim)
            elif block_size[d] == 1:  # singleton dimension
                nx_vals.append(1)
            else:  # normal case
                nx_vals.append(root_grid_size[d] * 2**level + 2 * num_ghost)
        nx1 = nx_vals[0]
        nx2 = nx_vals[1]
        nx3 = nx_vals[2]
        lx1 = nx1 // block_size[0]
        lx2 = nx2 // block_size[1]
        lx3 = nx3 // block_size[2]
        num_extended_dims = 0
        for nx in nx_vals:
            if nx > 1:
                num_extended_dims += 1

        # Set volume function for preset coordinates if needed
        coord = f.attrs['Coordinates'].decode('ascii', 'replace')
        if level < max_level and not subsample and not fast_restrict and vol_func is None:
            x1_rat = f.attrs['RootGridX1'][2]
            x2_rat = f.attrs['RootGridX2'][2]
            x3_rat = f.attrs['RootGridX3'][2]
            if (coord == 'cartesian' or coord == 'minkowski' or coord == 'tilted'
                    or coord == 'sinusoidal'):
                if ((nx1 == 1 or x1_rat == 1.0) and (nx2 == 1 or x2_rat == 1.0)
                        and (nx3 == 1 or x3_rat == 1.0)):
                    fast_restrict = True
                else:
                    def vol_func(xm, xp, ym, yp, zm, zp):
                        return (xp-xm) * (yp-ym) * (zp-zm)
            elif coord == 'cylindrical':
                if (nx1 == 1 and (nx2 == 1 or x2_rat == 1.0)
                        and (nx3 == 1 or x3_rat == 1.0)):
                    fast_restrict = True
                else:
                    def vol_func(rm, rp, phim, phip, zm, zp):
                        return (rp**2-rm**2) * (phip-phim) * (zp-zm)
            elif coord == 'spherical_polar' or coord == 'schwarzschild':
                if nx1 == 1 and nx2 == 1 and (nx3 == 1 or x3_rat == 1.0):
                    fast_restrict = True
                else:
                    def vol_func(rm, rp, thetam, thetap, phim, phip):
                        return ((rp**3-rm**3) * abs(np.cos(thetam)-np.cos(thetap))
                                * (phip-phim))
            elif coord == 'kerr-schild':
                if nx1 == 1 and nx2 == 1 and (nx3 == 1 or x3_rat == 1.0):
                    fast_restrict = True
                else:
                    a = vol_params[0]

                    def vol_func(rm, rp, thetam, thetap, phim, phip):
                        cosm = np.cos(thetam)
                        cosp = np.cos(thetap)
                        return (((rp**3-rm**3) * abs(cosm-cosp)
                                 + a**2 * (rp-rm) * abs(cosm**3-cosp**3)) * (phip-phim))
            else:
                raise AthenaError('Coordinates not recognized')

        # Set cell center functions for preset coordinates
        if center_func_1 is None:
            if (coord == 'cartesian' or coord == 'minkowski' or coord == 'tilted'
                    or coord == 'sinusoidal' or coord == 'kerr-schild'):
                def center_func_1(xm, xp):
                    return 0.5 * (xm+xp)
            elif coord == 'cylindrical':
                def center_func_1(xm, xp):
                    return 2.0/3.0 * (xp**3-xm**3) / (xp**2-xm**2)
            elif coord == 'spherical_polar':
                def center_func_1(xm, xp):
                    return 3.0/4.0 * (xp**4-xm**4) / (xp**3-xm**3)
            elif coord == 'schwarzschild':
                def center_func_1(xm, xp):
                    return (0.5*(xm**3+xp**3)) ** (1.0/3.0)
            else:
                raise AthenaError('Coordinates not recognized')
        if center_func_2 is None:
            if (coord == 'cartesian' or coord == 'cylindrical' or coord == 'minkowski'
                    or coord == 'tilted' or coord == 'sinusoidal'
                    or coord == 'kerr-schild'):
                def center_func_2(xm, xp):
                    return 0.5 * (xm+xp)
            elif coord == 'spherical_polar':
                def center_func_2(xm, xp):
                    sm = np.sin(xm)
                    cm = np.cos(xm)
                    sp = np.sin(xp)
                    cp = np.cos(xp)
                    return (sp-xp*cp - sm+xm*cm) / (cm - cp)
            elif coord == 'schwarzschild':
                def center_func_2(xm, xp):
                    return np.arccos(0.5 * (np.cos(xm) + np.cos(xp)))
            else:
                raise AthenaError('Coordinates not recognized')
        if center_func_3 is None:
            if (coord == 'cartesian' or coord == 'cylindrical' or coord == 'tilted'
                    or coord == 'spherical_polar' or coord == 'minkowski'
                    or coord == 'sinusoidal' or coord == 'schwarzschild'
                    or coord == 'kerr-schild'):

                def center_func_3(xm, xp):
                    return 0.5 * (xm+xp)
            else:
                raise AthenaError('Coordinates not recognized')

        # Check output level compared to max level in file
        if level < max_level and not subsample and not fast_restrict:
            warnings.warn('Exact restriction being used: performance severely affected;'
                          + ' see documentation', AthenaWarning)
            sys.stderr.flush()
        if level > max_level:
            warnings.warn('Requested refinement level higher than maximum level in file:'
                          + ' all cells will be prolongated', AthenaWarning)
            sys.stderr.flush()

        # Check that subsampling and/or fast restriction will work if needed
        if level < max_level and (subsample or fast_restrict):
            if num_ghost > 0:
                raise AthenaError('Subsampling and fast restriction incompatible with'
                                  + ' ghost zones')
            max_restrict_factor = 2**(max_level-level)
            for current_block_size in block_size:
                if (current_block_size != 1
                        and current_block_size % max_restrict_factor != 0):
                    raise AthenaError('Block boundaries at finest level must be cell'
                                      + ' boundaries at desired level for subsampling or'
                                      + ' fast restriction to work')

        # Create list of all quantities if none given
        var_quantities = np.array([x.decode('ascii', 'replace')
                                   for x in f.attrs['VariableNames'][:]])
        coord_quantities = ('x1f', 'x2f', 'x3f', 'x1v', 'x2v', 'x3v')
        attr_quantities = [key for key in f.attrs]
        other_quantities = ('Levels',)
        if not new_data:
            quantities = data.values()
        elif quantities is None:
            quantities = var_quantities
        else:
            for q in quantities:
                if q not in var_quantities and q not in coord_quantities:
                    possibilities = '", "'.join(var_quantities)
                    possibilities = '"' + possibilities + '"'
                    error_string = ('Quantity not recognized: file does not include "{0}"'
                                    + ' but does include {1}')
                    raise AthenaError(error_string.format(q, possibilities))
        quantities = [str(q) for q in quantities if q not in coord_quantities
                      and q not in attr_quantities and q not in other_quantities]

        # Store file attribute metadata
        for key in attr_quantities:
            data[str(key)] = f.attrs[key]

        # Get metadata describing file layout
        num_blocks = f.attrs['NumMeshBlocks']
        dataset_names = np.array([x.decode('ascii', 'replace')
                                  for x in f.attrs['DatasetNames'][:]])
        dataset_sizes = f.attrs['NumVariables'][:]
        dataset_sizes_cumulative = np.cumsum(dataset_sizes)
        variable_names = np.array([x.decode('ascii', 'replace')
                                   for x in f.attrs['VariableNames'][:]])
        quantity_datasets = []
        quantity_indices = []
        for q in quantities:
            var_num = np.where(variable_names == q)[0][0]
            dataset_num = np.where(dataset_sizes_cumulative > var_num)[0][0]
            if dataset_num == 0:
                dataset_index = var_num
            else:
                dataset_index = var_num - dataset_sizes_cumulative[dataset_num-1]
            quantity_datasets.append(dataset_names[dataset_num])
            quantity_indices.append(dataset_index)

        # Locate fine block for coordinates in case of slice
        fine_block = np.where(levels == max_level)[0][0]
        x1m = f['x1f'][fine_block, 0]
        x1p = f['x1f'][fine_block, 1]
        x2m = f['x2f'][fine_block, 0]
        x2p = f['x2f'][fine_block, 1]
        x3m = f['x3f'][fine_block, 0]
        x3p = f['x3f'][fine_block, 1]

        # Populate coordinate arrays
        face_funcs = (face_func_1, face_func_2, face_func_3)
        center_funcs = (center_func_1, center_func_2, center_func_3)
        for d, nx, face_func, center_func in zip(range(1, 4), nx_vals, face_funcs,
                                                 center_funcs):
            xf = 'x' + repr(d) + 'f'
            xv = 'x' + repr(d) + 'v'
            if nx == 1:
                xm = (x1m, x2m, x3m)[d-1]
                xp = (x1p, x2p, x3p)[d-1]
                data[xf] = np.array([xm, xp])
            else:
                xmin = f.attrs['RootGridX' + repr(d)][0]
                xmax = f.attrs['RootGridX' + repr(d)][1]
                xrat_root = f.attrs['RootGridX' + repr(d)][2]
                if xrat_root == -1.0 and face_func is None:
                    raise AthenaError('Must specify user-defined face_func_{0}'.format(d))
                elif face_func is not None:
                    if num_ghost > 0:
                        raise AthenaError('Ghost zones incompatible with user-defined'
                                          + ' coordinate spacing')
                    data[xf] = face_func(xmin, xmax, xrat_root, nx+1)
                elif xrat_root == 1.0:
                    if np.all(levels == level):
                        data[xf] = np.empty(nx + 1)
                        for n_block in range(int((nx - 2*num_ghost)
                                                 // (block_size[d-1] - 2*num_ghost))):
                            sample_block = np.where(logical_locations[:, d-1]
                                                    == n_block)[0][0]
                            index_low = n_block * (block_size[d-1] - 2*num_ghost)
                            index_high = index_low + block_size[d-1] + 1
                            data[xf][index_low:index_high] = f[xf][sample_block, :]
                    else:
                        if num_ghost > 0:
                            raise AthenaError('Cannot use ghost zones with different'
                                              + ' refinement levels')
                        data[xf] = np.linspace(xmin, xmax, nx + 1)
                else:
                    if num_ghost > 0:
                        raise AthenaError('Ghost zones incompatible with non-uniform'
                                          + ' coordinate spacing')
                    xrat = xrat_root ** (1.0 / 2**level)
                    data[xf] = (xmin + (1.0-xrat**np.arange(nx+1))
                                / (1.0-xrat**nx) * (xmax-xmin))
            data[xv] = np.empty(nx)
            for i in range(nx):
                data[xv][i] = center_func(data[xf][i], data[xf][i+1])

        # Account for selection
        x1_select = False
        x2_select = False
        x3_select = False
        i_min = j_min = k_min = 0
        i_max = nx1
        j_max = nx2
        k_max = nx3
        error_string = '{0} must be {1} than {2} in order to intersect data range'
        if x1_min is not None and x1_min >= data['x1f'][1]:
            if x1_min >= data['x1f'][-1]:
                raise AthenaError(error_string.format('x1_min', 'less', data['x1f'][-1]))
            x1_select = True
            i_min = np.where(data['x1f'] <= x1_min)[0][-1]
        if x1_max is not None and x1_max <= data['x1f'][-2]:
            if x1_max <= data['x1f'][0]:
                raise AthenaError(error_string.format('x1_max', 'greater',
                                                      data['x1f'][0]))
            x1_select = True
            i_max = np.where(data['x1f'] >= x1_max)[0][0]
        if x2_min is not None and x2_min >= data['x2f'][1]:
            if x2_min >= data['x2f'][-1]:
                raise AthenaError(error_string.format('x2_min', 'less', data['x2f'][-1]))
            x2_select = True
            j_min = np.where(data['x2f'] <= x2_min)[0][-1]
        if x2_max is not None and x2_max <= data['x2f'][-2]:
            if x2_max <= data['x2f'][0]:
                raise AthenaError(error_string.format('x2_max', 'greater',
                                                      data['x2f'][0]))
            x2_select = True
            j_max = np.where(data['x2f'] >= x2_max)[0][0]
        if x3_min is not None and x3_min >= data['x3f'][1]:
            if x3_min >= data['x3f'][-1]:
                raise AthenaError(error_string.format('x3_min', 'less', data['x3f'][-1]))
            x3_select = True
            k_min = np.where(data['x3f'] <= x3_min)[0][-1]
        if x3_max is not None and x3_max <= data['x3f'][-2]:
            if x3_max <= data['x3f'][0]:
                raise AthenaError(error_string.format('x3_max', 'greater',
                                                      data['x3f'][0]))
            x3_select = True
            k_max = np.where(data['x3f'] >= x3_max)[0][0]
        if (x1_select or x2_select or x3_select) and num_ghost > 0:
            raise AthenaError('Cannot take subsets with ghost zones')

        # Adjust coordinates if selection made
        if x1_select:
            data['x1f'] = data['x1f'][i_min:i_max+1]
            data['x1v'] = data['x1v'][i_min:i_max]
        if x2_select:
            data['x2f'] = data['x2f'][j_min:j_max+1]
            data['x2v'] = data['x2v'][j_min:j_max]
        if x3_select:
            data['x3f'] = data['x3f'][k_min:k_max+1]
            data['x3v'] = data['x3v'][k_min:k_max]

        # Prepare arrays for data and bookkeeping
        if new_data:
            for q in quantities:
                data[q] = np.zeros((k_max-k_min, j_max-j_min, i_max-i_min), dtype=dtype)
            if return_levels:
                data['Levels'] = np.empty((k_max-k_min, j_max-j_min, i_max-i_min),
                                          dtype=np.int32)
        else:
            for q in quantities:
                data[q].fill(0.0)
        if not subsample and not fast_restrict and max_level > level:
            restricted_data = np.zeros((lx3, lx2, lx1), dtype=bool)

        # Go through blocks in data file
        for block_num in range(num_blocks):
            # Extract location information
            block_level = levels[block_num]
            block_location = logical_locations[block_num, :]

            # Prolongate coarse data and copy same-level data
            if block_level <= level:
                # Calculate scale (number of copies per dimension)
                s = 2 ** (level - block_level)

                # Calculate destination indices, without selection
                il_d = (block_location[0] * (block_size[0] - 2*num_ghost) * s
                        if nx1 > 1 else 0)
                jl_d = (block_location[1] * (block_size[1] - 2*num_ghost) * s
                        if nx2 > 1 else 0)
                kl_d = (block_location[2] * (block_size[2] - 2*num_ghost) * s
                        if nx3 > 1 else 0)
                iu_d = il_d + block_size[0] * s if nx1 > 1 else 1
                ju_d = jl_d + block_size[1] * s if nx2 > 1 else 1
                ku_d = kl_d + block_size[2] * s if nx3 > 1 else 1

                # Calculate (prolongated) source indices, with selection
                il_s = max(il_d, i_min) - il_d
                jl_s = max(jl_d, j_min) - jl_d
                kl_s = max(kl_d, k_min) - kl_d
                iu_s = min(iu_d, i_max) - il_d
                ju_s = min(ju_d, j_max) - jl_d
                ku_s = min(ku_d, k_max) - kl_d
                if il_s >= iu_s or jl_s >= ju_s or kl_s >= ku_s:
                    continue

                # Account for selection in destination indices
                il_d = max(il_d, i_min) - i_min
                jl_d = max(jl_d, j_min) - j_min
                kl_d = max(kl_d, k_min) - k_min
                iu_d = min(iu_d, i_max) - i_min
                ju_d = min(ju_d, j_max) - j_min
                ku_d = min(ku_d, k_max) - k_min

                # Assign values
                for q, dataset, index in zip(quantities, quantity_datasets,
                                             quantity_indices):
                    block_data = f[dataset][index, block_num, :]
                    if s > 1:
                        if nx1 > 1:
                            block_data = np.repeat(block_data, s, axis=2)[:, :, il_s:iu_s]
                        if nx2 > 1:
                            block_data = np.repeat(block_data, s, axis=1)[:, jl_s:ju_s, :]
                        if nx3 > 1:
                            block_data = np.repeat(block_data, s, axis=0)[kl_s:ku_s, :, :]
                        data[q][kl_d:ku_d, jl_d:ju_d, il_d:iu_d] = block_data
                    else:
                        data[q][kl_d:ku_d, jl_d:ju_d, il_d:iu_d] = block_data[kl_s:ku_s,
                                                                              jl_s:ju_s,
                                                                              il_s:iu_s]

            # Restrict fine data
            else:
                # Calculate scale
                s = 2 ** (block_level - level)

                # Calculate destination indices, without selection
                il_d = block_location[0] * block_size[0] // s if nx1 > 1 else 0
                jl_d = block_location[1] * block_size[1] // s if nx2 > 1 else 0
                kl_d = block_location[2] * block_size[2] // s if nx3 > 1 else 0
                iu_d = il_d + block_size[0] // s if nx1 > 1 else 1
                ju_d = jl_d + block_size[1] // s if nx2 > 1 else 1
                ku_d = kl_d + block_size[2] // s if nx3 > 1 else 1

                # Calculate (restricted) source indices, with selection
                il_s = max(il_d, i_min) - il_d
                jl_s = max(jl_d, j_min) - jl_d
                kl_s = max(kl_d, k_min) - kl_d
                iu_s = min(iu_d, i_max) - il_d
                ju_s = min(ju_d, j_max) - jl_d
                ku_s = min(ku_d, k_max) - kl_d
                if il_s >= iu_s or jl_s >= ju_s or kl_s >= ku_s:
                    continue

                # Account for selection in destination indices
                il_d = max(il_d, i_min) - i_min
                jl_d = max(jl_d, j_min) - j_min
                kl_d = max(kl_d, k_min) - k_min
                iu_d = min(iu_d, i_max) - i_min
                ju_d = min(ju_d, j_max) - j_min
                ku_d = min(ku_d, k_max) - k_min

                # Account for restriction in source indices
                if nx1 > 1:
                    il_s *= s
                    iu_s *= s
                if nx2 > 1:
                    jl_s *= s
                    ju_s *= s
                if nx3 > 1:
                    kl_s *= s
                    ku_s *= s

                # Apply subsampling
                if subsample:
                    # Calculate fine-level offsets (nearest cell at or below center)
                    o1 = s//2 - 1 if nx1 > 1 else 0
                    o2 = s//2 - 1 if nx2 > 1 else 0
                    o3 = s//2 - 1 if nx3 > 1 else 0

                    # Assign values
                    for q, dataset, index in zip(quantities, quantity_datasets,
                                                 quantity_indices):
                        data[q][kl_d:ku_d,
                                jl_d:ju_d,
                                il_d:iu_d] = f[dataset][index, block_num, kl_s+o3:ku_s:s,
                                                        jl_s+o2:ju_s:s, il_s+o1:iu_s:s]

                # Apply fast (uniform Cartesian) restriction
                elif fast_restrict:
                    # Calculate fine-level offsets
                    io_vals = range(s) if nx1 > 1 else (0,)
                    jo_vals = range(s) if nx2 > 1 else (0,)
                    ko_vals = range(s) if nx3 > 1 else (0,)

                    # Assign values
                    for q, dataset, index in zip(quantities, quantity_datasets,
                                                 quantity_indices):
                        for ko in ko_vals:
                            for jo in jo_vals:
                                for io in io_vals:
                                    data[q][kl_d:ku_d,
                                            jl_d:ju_d,
                                            il_d:iu_d] += f[dataset][index, block_num,
                                                                     kl_s+ko:ku_s:s,
                                                                     jl_s+jo:ju_s:s,
                                                                     il_s+io:iu_s:s]
                        data[q][kl_d:ku_d, jl_d:ju_d, il_d:iu_d] /= s ** num_extended_dims

                # Apply exact (volume-weighted) restriction
                else:
                    # Calculate sets of indices
                    i_s_vals = range(il_s, iu_s)
                    j_s_vals = range(jl_s, ju_s)
                    k_s_vals = range(kl_s, ku_s)
                    i_d_vals = range(il_d, iu_d)
                    j_d_vals = range(jl_d, ju_d)
                    k_d_vals = range(kl_d, ku_d)
                    if nx1 > 1:
                        i_d_vals = np.repeat(i_d_vals, s)
                    if nx2 > 1:
                        j_d_vals = np.repeat(j_d_vals, s)
                    if nx3 > 1:
                        k_d_vals = np.repeat(k_d_vals, s)

                    # Accumulate values
                    for k_s, k_d in zip(k_s_vals, k_d_vals):
                        if nx3 > 1:
                            x3m = f['x3f'][block_num, k_s]
                            x3p = f['x3f'][block_num, k_s+1]
                        for j_s, j_d in zip(j_s_vals, j_d_vals):
                            if nx2 > 1:
                                x2m = f['x2f'][block_num, j_s]
                                x2p = f['x2f'][block_num, j_s+1]
                            for i_s, i_d in zip(i_s_vals, i_d_vals):
                                if nx1 > 1:
                                    x1m = f['x1f'][block_num, i_s]
                                    x1p = f['x1f'][block_num, i_s+1]
                                vol = vol_func(x1m, x1p, x2m, x2p, x3m, x3p)
                                for q, dataset, index in zip(quantities,
                                                             quantity_datasets,
                                                             quantity_indices):
                                    data[q][k_d, j_d, i_d] += vol * f[dataset][index,
                                                                               block_num,
                                                                               k_s, j_s,
                                                                               i_s]
                    loc1 = (nx1 > 1) * block_location[0] // s
                    loc2 = (nx2 > 1) * block_location[1] // s
                    loc3 = (nx3 > 1) * block_location[2] // s
                    restricted_data[loc3, loc2, loc1] = True

            # Set level information for cells in this block
            if return_levels:
                data['Levels'][kl_d:ku_d, jl_d:ju_d, il_d:iu_d] = block_level

    # Remove volume factors from restricted data
    if level < max_level and not subsample and not fast_restrict:
        for loc3 in range(lx3):
            for loc2 in range(lx2):
                for loc1 in range(lx1):
                    if restricted_data[loc3, loc2, loc1]:
                        il = loc1 * block_size[0]
                        jl = loc2 * block_size[1]
                        kl = loc3 * block_size[2]
                        iu = il + block_size[0]
                        ju = jl + block_size[1]
                        ku = kl + block_size[2]
                        il = max(il, i_min) - i_min
                        jl = max(jl, j_min) - j_min
                        kl = max(kl, k_min) - k_min
                        iu = min(iu, i_max) - i_min
                        ju = min(ju, j_max) - j_min
                        ku = min(ku, k_max) - k_min
                        for k in range(kl, ku):
                            if nx3 > 1:
                                x3m = data['x3f'][k]
                                x3p = data['x3f'][k+1]
                            for j in range(jl, ju):
                                if nx2 > 1:
                                    x2m = data['x2f'][j]
                                    x2p = data['x2f'][j+1]
                                for i in range(il, iu):
                                    if nx1 > 1:
                                        x1m = data['x1f'][i]
                                        x1p = data['x1f'][i+1]
                                    vol = vol_func(x1m, x1p, x2m, x2p, x3m, x3p)
                                    for q in quantities:
                                        data[q][k, j, i] /= vol

    # Return dictionary containing requested data arrays
    if check_nan_flag:
        for key, val in data.items():
            if key in quantities:
                check_nan(val)

    return data


# ========================================================================================

def restrict_like(vals, levels, vols=None):
    """Average cell values according to given mesh refinement scheme."""

    # Determine maximum amount of restriction
    nx3, nx2, nx1 = vals.shape
    max_level = np.max(levels)
    if nx3 > 1 and nx3 % 2**max_level != 0:
        raise AthenaError('x3-dimension wrong size to be restricted')
    if nx2 > 1 and nx2 % 2**max_level != 0:
        raise AthenaError('x2-dimension wrong size to be restricted')
    if nx1 % 2**max_level != 0:
        raise AthenaError('x1-dimension wrong size to be restricted')

    # Construct volume weighting
    if vols is None:
        vols = np.ones_like(vals)
    else:
        if vols.shape != vals.shape:
            raise AthenaError('Array of volumes must match cell values in size')

    # Restrict data
    vals_restricted = np.copy(vals)
    for level in range(max_level):
        level_difference = max_level - level
        stride = 2 ** level_difference
        if nx3 > 1:
            vals_level = np.reshape(vals * vols, (nx3//stride, stride, nx2//stride,
                                                  stride, nx1//stride, stride))
            vols_level = np.reshape(vols, (nx3//stride, stride, nx2//stride, stride,
                                           nx1//stride, stride))
            vals_sum = np.sum(np.sum(np.sum(vals_level, axis=5), axis=3), axis=1)
            vols_sum = np.sum(np.sum(np.sum(vols_level, axis=5), axis=3), axis=1)
            vals_level = np.repeat(np.repeat(np.repeat(vals_sum / vols_sum, stride,
                                                       axis=0),
                                             stride, axis=1),
                                   stride, axis=2)
        elif nx2 > 1:
            vals_level = np.reshape(vals * vols, (nx2//stride, stride, nx1//stride,
                                                  stride))
            vols_level = np.reshape(vols, (nx2//stride, stride, nx1//stride, stride))
            vals_sum = np.sum(np.sum(vals_level, axis=3), axis=1)
            vols_sum = np.sum(np.sum(vols_level, axis=3), axis=1)
            vals_level = np.repeat(np.repeat(vals_sum / vols_sum, stride, axis=0),
                                   stride, axis=1)
        else:
            vals_level = np.reshape(vals * vols, (nx1//stride, stride))
            vols_level = np.reshape(vols, (nx1//stride, stride))
            vals_sum = np.sum(vals_level, axis=1)
            vols_sum = np.sum(vols_level, axis=1)
            vals_level = np.repeat(vals_sum / vols_sum, stride, axis=0)
        vals_restricted = np.where(levels == level, vals_level, vals_restricted)
    return vals_restricted


# ========================================================================================

def athinput(filename):
    """Read athinput file and returns a dictionary of dictionaries."""

    # Read data
    with open(filename, 'r') as athinput:
        # remove comments, extra whitespace, and empty lines
        lines = filter(None, [i.split('#')[0].strip() for i in athinput.readlines()])
    data = {}
    # split into blocks, first element will be empty
    blocks = ('\n'.join(lines)).split('<')[1:]

    # Function for interpreting strings numerically
    def typecast(x):
        if '_' in x:
            return x
        try:
            return int(x)
        except ValueError:
            pass
        try:
            return float(x)
        except ValueError:
            pass
        try:
            return complex(x)
        except ValueError:
            pass
        return x

    # Function for parsing assignment based on first '='
    def parse_line(line):
        out = [i.strip() for i in line.split('=')]
        out[1] = '='.join(out[1:])
        out[1] = typecast(out[1])
        return out[:2]

    # Assign values into dictionaries
    for block in blocks:
        info = list(filter(None, block.split('\n')))
        key = info.pop(0)[:-1]  # last character is '>'
        data[key] = dict(map(parse_line, info))
    return data


# ========================================================================================

class AthenaError(RuntimeError):
    """General exception class for Athena++ read functions."""
    pass


class AthenaWarning(RuntimeWarning):
    """General warning class for Athena++ read functions."""
    pass



# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.environment import dHybridRtemplate
from kgsim.simulation import SimulationGroup
from kgsim.dhybridr.initializer import dHybridRinitializer, dHybridRconfig, dHybridRSnapshot
from kgsim.dhybridr.dhybridr import dHybridR

from kbasic.parsing import File, Folder
from kbasic.Tex import texfraction
from typing import Iterable

import numpy as np
from matplotlib.pyplot import cm

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
def field_dot(A: np.ndarray, B: np.ndarray) -> np.ndarray: return np.sum(A * B, axis=0)

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class TurbInit(dHybridRinitializer):
    def __init__(
        self,
        simulation
    ):
        self.simulation = simulation
        self.dims = len(self.simulation.input.boxsize)
        self.config = dHybridRconfig(simulation)
        self.mach = self.config.mach
        self.simulation.mach = self.mach 
        self.dB = self.config.dB
        self.simulation.dB = self.dB
        self.amplitude: tuple = (self.dB, self.mach)
        #this gives us L, N's, shape, and d's, as well as the basic code to produce B and u fields and save those as d files
        dHybridRinitializer.__init__(self, simulation)
        self.configure()
    def configure(self):
        #config works different for simulations of different dimensions
        match self.dims:
            case 1: print("not implemented, low-key not sure you can do this in dHybridR????")
            case 2:
                #set the initial k space annuli for producing turbulence
                if "kinit" not in self.config.params: self.kinit = (1, np.pi), (1, np.pi) #Default value if config file has no kinit
                else: #if the config file has a kinit value use that
                    if len(self.config.kinit)==2: self.kinit = self.config.kinit, self.config.kinit 
                    elif len(self.config.kinit)==4: self.kinit = self.config.kinit[:2], self.config.kinit[2:]
                    else: raise ValueError(f"config file's kinit value is invalid\nkinit={self.config.kinit}")
                    #set the simulations kinit value for future use
                self.simulation.kinit = self.kinit

                #set k range 
                self.kmin = 2 * np.pi / max(self.L)
                self.kmax = 2 * np.pi / min([self.dx, self.dy]) #is this even used for anything? I don't think so?

                #set k vectors and compute their magnitude
                self.k = np.mgrid[
                    -self.Ny // 2: self.Ny // 2,
                    -self.Nx // 2: self.Nx // 2
                ][::-1] * self.kmin
                self.kmag = np.hypot(*self.k)
                self.kmag[self.kmag==0] = np.nan

                if not self.simulation.compressed:
                    l = self.input.niter if not self.simulation.output.exists else len(self.simulation.B)*self.input.ndump
                    self.simulation.time = np.arange(0, l, self.input.ndump) * self.input.dt
                    self.simulation.tau = self.simulation.time * max(self.mach if isinstance(self.mach, Iterable) else [self.mach]) / (max(self.input.boxsize))
                elif self.simulation.compressed:
                    self.simulation.time = np.array([int(x[-11:-3]) for x in self.simulation.density.file_names]) * self.simulation.input.dt
                    self.simulation.tau = self.simulation.time * max(self.mach if isinstance(self.mach, Iterable) else [self.mach]) / (max(self.input.boxsize))
                # if "peak_jz" in self.config.params: 
                #     self.simulation.peak_jz_ind = int(self.config.peak_jz)
                #     self.simulation.initial = dHybridRSnapshot(self.simulation,0)
                #     self.simulation.peak = dHybridRSnapshot(self.simulation, self.simulation.peak_jz_ind)
                #     self.simulation.snapshots = [
                #         self.simulation.initial, self.simulation.peak
                #     ]+[
                #         dHybridRSnapshot(self.simulation, np.argmin(abs(self.simulation.tau - n))) for n in range(1, int(self.simulation.tau[-1]//1))
                #     ]
            case 3:
                #set the initial k space annuli for producing turbulence
                if "kinit" not in self.config.params: self.kinit = (1, 2*np.pi), (1, 2*np.pi), (1, 2*np.pi) #Default value if config file has no kinit
                else: #if the config file has a kinit value use that
                    if len(self.config.kinit)==2: self.kinit = self.config.kinit, self.config.kinit, self.config.kinit
                    # if there are two sets then assume it goes perp, par -> kinit_perp, kinit_perp, kinit_par
                    elif len(self.config.kinit)==4: self.kinit = self.config.kinit[:2], self.config.kinit[:2], self.config.kinit[2:]
                    elif len(self.config.kinit)==6: self.kinit = self.config.kinit[:2], self.config.kinit[2:4], self.config.kinit[4:]
                    else: raise ValueError(f"config file's kinit value is invalid\nkinit={self.config.kinit}")
                #set the simulations kinit value for future use
                self.simulation.kinit = self.kinit
                #set k range 
                self.kmin = 2 * np.pi / max(self.L)
                self.kmax = 2 * np.pi / min([self.dx, self.dy, self.dz]) #is this even used for anything? I don't think so?

                #set k vectors and compute their magnitude
                self.k = np.mgrid[
                    -self.Ny // 2: self.Ny // 2,
                    -self.Nx // 2: self.Nx // 2,
                    -self.Nz // 2: self.Nz // 2
                ][::-1] * self.kmin
                self.kmag = np.sqrt(self.k[0]**2 + self.k[1]**2 + self.k[2]**2)
                self.kmag[self.kmag==0] = np.nan
                #set times
                if not self.simulation.compressed:
                    l = self.input.niter if not self.simulation.output.exists else len(self.simulation.B)*self.input.ndump
                    self.simulation.time = np.arange(0, l, self.input.ndump) * self.input.dt
                    self.simulation.tau = self.simulation.time * max(self.mach if isinstance(self.mach, Iterable) else [self.mach]) / (max(self.input.boxsize))
                elif self.simulation.compressed:
                    self.simulation.time = np.array([int(x[-11:-3]) for x in self.simulation.density.file_names]) * self.simulation.input.dt
                    self.simulation.tau = self.simulation.time * max(self.mach if isinstance(self.mach, Iterable) else [self.mach]) / (max(self.input.boxsize))
                if "peak_jz" in self.config.params: 
                    self.simulation.peak_jz_ind = int(self.config.peak_jz)
                    self.simulation.initial = dHybridRSnapshot(self.simulation,0)
                    self.simulation.peak = dHybridRSnapshot(self.simulation, self.simulation.peak_jz_ind)
                    self.simulation.snapshots = [
                    self.simulation.initial, self.simulation.peak
                    ]+[
                    dHybridRSnapshot(self.simulation, np.argmin(abs(self.simulation.tau - n))) for n in range(1, int(self.simulation.tau[-1]//1))
                    ]
    def fluctuate3D(self, field, amp, no_div=True):
        init_mask = np.array([
            np.where((self.kinit[0][0] * self.kmin < self.kmag)&(self.kmag < self.kinit[0][1]*self.kmin),True,False),
            np.where((self.kinit[1][0] * self.kmin < self.kmag)&(self.kmag < self.kinit[1][1]*self.kmin),True,False),
            np.where((self.kinit[2][0] * self.kmin < self.kmag)&(self.kmag < self.kinit[2][1]*self.kmin),True,False)
        ])
        M = np.sum(init_mask)
        phases = np.exp(2j * np.pi * np.random.random(field.shape))

        FT = np.zeros(field.shape, dtype=complex)
        FT[0][init_mask[0]] = amp[0] * np.pi / 2
        FT[1][init_mask[1]] = amp[1] * np.pi
        FT[2][init_mask[2]] = amp[2] * np.pi / 2
        FT *= phases
        # subtract off the parallel x/y components
        if no_div: FT -= field_dot(FT, self.k / self.kmag) * self.k / self.kmag
        FT[np.isnan(FT)] = 0
        # apply the condition to make this real
        _fx = np.roll(FT[1, ::-1, ::-1, ::-1], 1, axis=(0, 1, 2))
        FT[1, :self.Ny // 2, :self.Nx // 2, :self.Nz // 2] = np.conj(_fx[:self.Ny // 2, :self.Nx // 2, :self.Nz // 2])
        FT[1, self.Ny // 2:, :self.Nx // 2, :self.Nz // 2] = np.conj(_fx[self.Ny // 2:, :self.Nx // 2, :self.Nz // 2])
        FT[1, :self.Ny // 2, :self.Nx // 2, self.Nz // 2:] = np.conj(_fx[:self.Ny // 2, :self.Nx // 2, self.Nz // 2:])
        FT[1, self.Ny // 2:, :self.Nx // 2, self.Nz // 2:] = np.conj(_fx[self.Ny // 2:, :self.Nx // 2, self.Nz // 2:])
        _fy = np.roll(FT[0, ::-1, ::-1, ::-1], 1, axis=(0, 1, 2))
        FT[0, :self.Ny // 2, :self.Nx // 2, :self.Nz // 2] = np.conj(_fy[:self.Ny // 2, :self.Nx // 2, :self.Nz // 2])
        FT[0, :self.Ny // 2, :self.Nx // 2, :self.Nz // 2] = np.conj(_fy[:self.Ny // 2, :self.Nx // 2, :self.Nz // 2])
        FT[0, self.Ny // 2:, :self.Nx // 2, :self.Nz // 2] = np.conj(_fy[self.Ny // 2:, :self.Nx // 2, :self.Nz // 2])
        FT[0, :self.Ny // 2, :self.Nx // 2, self.Nz // 2:] = np.conj(_fy[:self.Ny // 2, :self.Nx // 2, self.Nz // 2:])
        FT[0, self.Ny // 2:, :self.Nx // 2, self.Nz // 2:] = np.conj(_fy[self.Ny // 2:, :self.Nx // 2, self.Nz // 2:])
        _fz = np.roll(FT[2, ::-1, ::-1, ::-1], 1, axis=(0, 1, 2))
        FT[2, :self.Ny // 2, :self.Nx // 2, :self.Nz // 2] = np.conj(_fz[:self.Ny // 2, :self.Nx // 2, :self.Nz // 2])
        FT[2, :self.Ny // 2, :self.Nx // 2, :self.Nz // 2] = np.conj(_fz[:self.Ny // 2, :self.Nx // 2, :self.Nz // 2])
        FT[2, self.Ny // 2:, :self.Nx // 2, :self.Nz // 2] = np.conj(_fz[self.Ny // 2:, :self.Nx // 2, :self.Nz // 2])
        FT[2, :self.Ny // 2, :self.Nx // 2, self.Nz // 2:] = np.conj(_fz[:self.Ny // 2, :self.Nx // 2, self.Nz // 2:])
        FT[2, self.Ny // 2:, :self.Nx // 2, self.Nz // 2:] = np.conj(_fz[self.Ny // 2:, :self.Nx // 2, self.Nz // 2:])

        self.FT = FT
        # take the inverse fourier transform
        y: np.ndarray = np.real(
            np.fft.ifftn(
                np.fft.ifftshift(
                    FT
                )
            )
        ) / M * self.Nx * self.Ny * self.Nz
        rms = np.sqrt(np.nanmean(y[0]**2 + y[1]**2 + y[2]**2))
        factor = np.array(amp / rms)
        y[0] = factor[0] * y[0]
        y[1] = factor[1] * y[1]
        y[2] = factor[2] * y[2]
        return np.float32(y)
    def fluctuate2D(self, field, amp, no_div=True):
        """
        Given the initialization create a 2d array the same shape as the simulation which will smoothly fluctuate
        over length scales kinit
        :return y: np.ndarray[np.float32]: random fluctuations set by parameters passed to __init__
        """
        init_mask: np.ndarray = np.where((self.kinit[0][0] * self.kmin < self.kmag) & (self.kmag < self.kinit[0][1] * self.kmin))
        M: int = len(init_mask[0])  # number of cells with an amplitude
        phases: np.ndarray = np.exp(2j * np.pi * np.random.random(field.shape))  # randomized complex phases
        phases[2] *= 0  # don't wiggle the z component
        # Setting the fourier transform
        FT: np.ndarray = np.zeros(field.shape, dtype=complex)  # same shape as field
        FT[0][init_mask] = amp * np.pi / 2 # set x and y amplitudes
        FT[1][init_mask] = amp * np.pi
        FT *= phases  # apply phases
        # subtract off the parallel x/y components
        if no_div:
            FT[:2] -= field_dot(FT[:2], self.k / self.kmag) * self.k / self.kmag
        FT[np.isnan(FT)] = 0
        # apply the condition to make this real
        _fx = np.roll(FT[1, ::-1, ::-1], 1, axis=(0, 1))
        FT[1, :self.Ny // 2] = np.conj(_fx[:self.Ny // 2])

        _fy = np.roll(FT[0, ::-1, ::-1], 1, axis=(0, 1))
        FT[0, :self.Ny // 2] = np.conj(_fy[:self.Ny // 2])

        # I think we have to fix the zero line
        FT[1, self.Ny // 2, 1:self.Nx // 2] = FT[1, self.Ny // 2, self.Nx // 2 + 1:][::-1]
        FT[1, self.Ny // 2, 1:self.Nx // 2] = np.conj(FT[1, self.Ny // 2, self.Nx // 2 + 1:][::-1])
        FT[1, self.Ny // 2, :] = 0.j
        FT[1, :, self.Nx // 2] = 0.j

        FT[0, self.Ny // 2, 1:self.Nx // 2] = FT[0, self.Ny // 2, self.Nx // 2 + 1:][::-1]
        FT[0, self.Ny // 2, 1:self.Nx // 2] = np.conj(FT[0, self.Ny // 2, self.Nx // 2 + 1:][::-1])
        
        self.FT = FT

        # take the inverse fourier transform
        y: np.ndarray = np.array([*np.real(
            np.fft.ifft2(
                np.fft.ifftshift(
                    FT[:2]
                )
            )
        ), np.zeros(FT[0].shape)]) / M * self.Nx * self.Ny

        rms = np.sqrt(np.nanmean(y[0]**2 + y[1]**2))
        y *= (amp / rms)
        return np.float32(y)   
    def fluctuate(self, field, amp, no_div=True):
        match self.dims:
            case 2: return self.fluctuate2D(field, amp, no_div=no_div)
            case 3: return self.fluctuate3D(field, amp, no_div=no_div)
    def construct_field(self, x, y, z, amp, no_div=True):
        """
        Constructs a 3 x N x N array representing a constant x, y, and z component with additional fluctuations
        :param x:
        :param y:
        :param z:
        :param no_div: whether or not to ensure that the divergence of the field is 0 when applying fluctuations
        :return field:
        """
        base_field = np.array([
            np.zeros(self.shape) + y,
            np.zeros(self.shape) + x,
            np.zeros(self.shape) + z
        ], dtype=np.float32)

        base_rms: float = np.sqrt(np.mean(base_field ** 2))
        fluctuations: np.ndarray = self.fluctuate(base_field, amp, no_div=no_div)
        field: np.ndarray = base_field + fluctuations
        alt_rms: float = np.sqrt(np.mean(base_field ** 2))
        if not any([base_rms == 0, alt_rms == 0]):
            field *= base_rms / alt_rms
        return field
    def build_B_field(self): self.B = self.construct_field(0, 0, 1, self.amplitude[0])
    def build_u_field(self): self.u = self.construct_field(0, 0, 0, self.amplitude[1])
class TurbSim(dHybridR):
    def __init__(
            self, 
            path: str,
            caching: bool = False,
            verbose: bool = False,
            template: Folder = dHybridRtemplate,
            compressed: bool = False
        ) -> None:
        dHybridR.__init__(self, path, caching=caching, verbose=verbose, template=template, compressed=compressed)
        self.config = dHybridRconfig(self, mode='turb')
        self.initializer = TurbInit(self)
class TurbGroup(SimulationGroup):
    def __init__(self, path, sort='mach', verbose=True, **sim_kwds):
        SimulationGroup.__init__(self, path, simtype=TurbSim, **sim_kwds)
        try: 
            self.sort_by(sort)
        except KeyError: 
            if verbose: print(f"could not sort by {sort}, using default order...")
    def colorer(self, cmap=cm.plasma): 
        machs = [x.mach for x in self.simulations.values()]
        return [cmap(np.log2(m/min(machs))/np.log2((max(machs)+.1)/min(machs))) for m in machs]
    def labeler(self): return [
        r"$\mathcal{M} = $"+f"{int(x.mach) if x.mach.is_integer() else texfraction(x.mach)}" for x in self.simulations.values()]


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
import numpy as np
from contextlib import contextmanager
from tqdm import tqdm
import inspect

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
def where_closest(arr:np.ndarray, x): return np.argmin(np.abs(arr-x))
def where_between(arr:np.ndarray, low, high): return np.where((arr>=low)&(arr<=high))
def bin_this(x, y, n_bins=50, func=np.nanmean):
    xbins = np.linspace(np.nanmin(x),np.nanmax(x),n_bins)
    Y,error = list(),list()
    for i,low_edge in enumerate(xbins[:-1]):
        high_edge = xbins[i+1]
        mask = (low_edge<x)&(x<high_edge)
        yin = y[mask]
        Y.append(func(yin))
        error.append(np.nanstd(yin)/np.sqrt(len(yin)))
    return np.array(xbins)[:-1],np.array(Y),np.array(error)
@contextmanager
def redirect_to_tqdm():
    # Store builtin print
    old_print = print
    def new_print(*args, **kwargs):
        # If tqdm.tqdm.write raises error, use builtin print
        try:
            tqdm.write(*args, **kwargs)
        except:
            old_print(*args, **kwargs)

    try:
        # Globaly replace print with new_print
        inspect.builtins.print = new_print
        yield
    finally:
        inspect.builtins.print = old_print
def progress_bar(iterator, **kwargs):
    with redirect_to_tqdm():
        for x in tqdm(iterator, **kwargs):
            yield x
def verbose_bar(iterator, verbose, **kwargs):
    return progress_bar(iterator, **kwargs) if verbose else iterator
def nan_clip(*args):
    mask = ~np.any([np.isnan(a) for a in args], axis=0)
    nanless_args = tuple([np.array(a)[mask] for a in args])
    return nanless_args
# Spectrum Functions
def kspec(image: np.ndarray) -> np.ndarray:
    return np.absolute(np.fft.fftshift(np.fft.fft2(image) / (1. * image.shape[0] * image.shape[1])))**2
def kspec1d(image: np.ndarray, bins: int = 100, return_bin_edges: bool = False) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    k = kspec(image)
    Ny, Nx = image.shape
    kmag = np.hypot(*np.mgrid[
                     -Ny // 2: Ny // 2,
                     -Nx // 2: Nx // 2
                     ][::-1])
    kmin = np.nanmin(kmag[kmag != 0])
    kmax = np.nanmax(kmag)
    kgrid = np.r_[[0], np.logspace(np.log10(3*kmin), np.log10(kmax), bins)]
    kx = np.array([np.mean([kgrid[i], kgrid[i+1]]) for i in range(len(kgrid)-1)])
    ks = np.array([np.mean(k[np.where((kgrid[i] < kmag) & (kmag < kgrid[i+1]))]) for i in range(len(kgrid)-1)])
    kerr = np.array([np.std(in_bin:=k[np.where((kgrid[i] < kmag) & (kmag < kgrid[i+1]))])/np.sqrt(len(in_bin)) for i in range(len(kgrid)-1)])
    if not return_bin_edges: return kx, ks, kerr
    return kgrid, kx, ks, kerr



# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.environment import dHybridRtemplate, frameDir
from kgsim.fields import ScalarField, VectorField
from kgsim.simulation import GenericSimulation, SimulationGroup
from kgsim.particles import Species, Particle
from kgsim.dhybridr.io import dHybridRinput, dHybridRout
from kgsim.dhybridr.initializer import dHybridRinitializer, TurbInit, dHybridRconfig
from kgsim.dhybridr.anvil_submit import AnvilSubmitScript

from kplot import show, func_video
from kbasic.Tex import texfraction
from kbasic.parsing import Folder
from kbasic.user_input import yesno
from os import system
from os.path import isdir
from glob import glob

from matplotlib.pyplot import cm as cmaps
import numpy as np 
from h5py import File as h5File

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                           Definitions                           <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
track_keys = ['B1', 'B2', 'B3', 'E1', 'E2', 'E3', 'ene', 'n', 'p1', 'p2', 'p3', 'q', 't', 'x1', 'x2', 'x3']

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# simulation parsing
def extract_energy(file_name: str) -> tuple:
    with h5File(file_name, 'r') as file:
        fE = np.mean(file["DATA"], axis=1)
        [low, high] = file["AXIS"]["X2 AXIS"][:]
        lne = np.linspace(low, high, fE.shape[0])
        dlne = np.diff(lne)[0]
        E = np.exp(lne)
        return E, fE, dlne
def iters(simulation) -> list[int]:
    return [int(fn[-11:-3]) for fn in simulation.density.file_names]
def times(simulation) -> list[float]:
    return list(np.array(iters(simulation)).astype(float) * simulation.dt)
def particle_video(
    sim,
    particles: list[str] | int,
    fname: str,
    background = None,      
    resume: bool = False,
    res: int = None,
    zfill: int = 10,
    dpi: int = 250,
    # paticle plotting keywords
    color = 'red',
    marker='.',
    ms=10, 
    # assume the rest are keywords for the show function
    **kwds
):
    # check species
    assert sim.input.num_species == 1, "only one species is implemented rn :-("
    # check the time resolution is valid
    if not res: res = sim.input.sp01.track_nstore
    dnp = sim.input.sp01.track_nstore
    dnb = sim.input.ndump
    assert (dnp % res == 0) & (res % dnb == 0), "your resolution must be an integer multiple of track_nstore and ndump must be an integer multiple of res."
    # handle the resuming
    if not resume: 
        system(f'rm {frameDir.path}/*')
        i_start = 0
    else:
        last_file = sorted(frameDir.children)[-1]
        last_iter = int(last_file[:-4])
        i_start = last_iter + res
    # find the file
    fn = sim.path+"/Output/Tracks/Sp01/track_Sp01.h5"
    with h5File(fn) as file:
        # extract particle tracks based on the particles argument
        tags = np.array(list(file.keys()))
        match particles:
            case [str(x), ]: selected = particles 
            case int(x): selected = np.random.choice(tags, size=x, replace=False)
        sx, sy = np.array([file[t]['x1'] for t in selected]).T, np.array([file[t]['x2'] for t in selected]).T 
        # setup background
        if not background: background = sim.density
        # initialize plots
        fig,ax,img = show(
            background[0], 
            x=range(0, sim.input.boxsize[0], sim.dx), y=range(0, sim.input.boxsize[1], sim.dy), 
            zorder=1,
            **kwds
        )
        line, = ax.plot(
            sx[0], sy[0], 
            ls='None', color=color, marker=marker, ms=ms,
            zorder = 2
        )
        # execute the loop
        def update(i: int, dnp=dnp, dnb=dnb):
            pind = i * dnp
            line.set_data(sx[pind], sy[pind])
            if pind % dnb == 0:
                bind = pind // dnb 
                img.set_array(background[bind])
        func_video(fname, fig, update, )

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class dHybridRparticle(Particle):
    def __init__(self, tag: str, species: Species, load: bool = False):
        self.tag = tag
        self.species = self.sp = species
        if load: self.load()
    def load(self):
        with h5File(self.species.path) as file:
            for k in track_keys:
                try: setattr(self, k, file[self.tag][k][:])
                except: continue
            self.x = self.x1 
            self.y = self.x2
    def follow_video(self, background=None, window=100, **kwds):
        sim = self.species.parent
        if not background: background = sim.density
        di = sim.input.sp01.track_nstore
        db = sim.input.ndump
        center = (window//2, window//2)
        fig,ax,img = show(background[0], **kwds)
        line, = ax.plot(*center, ls='None', marker='.', ms=10, color='r')
        for i in range(0, sim.input.niter, di):
            pind = i // di 
            img.set
class dHybridRspecies(Species):
    def __init__(self, species_number: int, parent):
        self.n = self.number = self.sp_num = species_number
        self.nstr = str(self.n).zfill(2)
        self.parent = parent 
        self.input = getattr(parent.input, f"sp{str(self.n).zfill(2)}")
        self.loaded = False
        if self.input.track_dump:
            self.path = parent.path+f"/Output/Tracks/Sp{self.nstr}/track_Sp{self.nstr}.h5"
            self.load()
    
    def load(self):
        with h5File(self.path) as file:
            self.tags = np.array(list(file.keys()))
            self.loaded = True 

    def sample(self, N, load=False): return np.array([dHybridRparticle(t, self) for t in np.random.choice(self.tags, size=N, load=load)])
class dHybridR(GenericSimulation):
    """
    A simulation class to interact with dHybridR simulations in python
    """
    def __init__(
            self, 
            path: str,
            caching: bool = False,
            verbose: bool = False,
            template: Folder = dHybridRtemplate,
            compressed: bool = False
        ) -> None:
        self.compressed = compressed
        #setup simulation
        GenericSimulation.__init__(self, path, caching=caching, verbose=verbose, template=template)
        #setup input, output, and restart folders
        self.parse_input()
        self.output = Folder(self.path+"/Output")
        if not self.output.exists and verbose:
            if yesno("There is no output, would you like to run this simulation?\n"): 
                self.run()
        elif self.output.exists: 
            self.parse_output()
            self.ncores: int = int(np.prod(self.input.node_number))
            self.ncores_charged: int = self.ncores + self.ncores % 128
            if self.out.exists:
                self.runtime: float = self.out.runtime #run time as calculated from out file, in hours
                self.corehours: float = self.runtime * np.prod(self.ncores)
                self.corehours_charged: float = self.runtime * self.ncores_charged
        self.restartDir = Folder(self.path+"/Restart")
    def __repr__(self) -> str: return self.name
    def __len__(self) -> int:
        if self.output.exists: 
            return len(self.density)
        else: return 0
    def create(self) -> None:
        self.template.copy(self.path)
        system(f"chmod 755 {self.path}/dHybridR")
    def parse_input(self) -> None:
        self.input = dHybridRinput(self.path+"/input/input")
        self.dt = self.input.dt
        self.niter = self.input.niter
        self.dx: float = self.input.boxsize[0]/self.input.ncells[0]
        self.dy: float = self.input.boxsize[1]/self.input.ncells[1]
        self.dz: float = self.input.boxsize[2]/self.input.ncells[2] if len(self.input.boxsize)==3 else np.inf
    def run(self, initializer: dHybridRinitializer, submit_script: AnvilSubmitScript) -> None:
        initializer.prepare_simulation()
        submit_script.write()
        system(f"sh {submit_script.path}")
    def parse_output(self) -> None:
        self.out = dHybridRout(self.path+"/out")
        kwargs = {'caching':self.caching, 'verbose':self.verbose, 'parent':self}
        self.B       = VectorField(self.path + "/Output/Fields/Magnetic/Total/", name="magnetic", latex="B", **kwargs)
        self.E       = VectorField(self.path + "/Output/Fields/Electric/Total/", name="electric", latex="E", **kwargs)
        self.etx1    = ScalarField(self.path + "/Output/Phase/etx1/Sp01/", **kwargs)
        self.pxx1    = ScalarField(self.path + "/Output/Phase/p1x1/Sp01/", **kwargs)
        self.pyx1    = ScalarField(self.path + "/Output/Phase/p2x1/Sp01/", **kwargs)
        self.pzx1    = ScalarField(self.path + "/Output/Phase/p3x1/Sp01/", **kwargs)
        self.density = ScalarField(self.path + "/Output/Phase/x3x2x1/Sp01/", name="density", latex=r"$\rho$", **kwargs)
        self.Pxx     = ScalarField(self.path + "/Output/Phase/PressureTen/Sp01/xx/", **kwargs)
        self.Pyy     = ScalarField(self.path + "/Output/Phase/PressureTen/Sp01/yy/", **kwargs)
        self.Pzz     = ScalarField(self.path + "/Output/Phase/PressureTen/Sp01/zz/", **kwargs)
        self.u       = VectorField(self.path + "/Output/Phase/FluidVel/Sp01/", name="bulkflow", latex="u", **kwargs)
        [
            self.energy_grid,
            self.energy_pdf,
            self.dlne
        ] = np.array([[*extract_energy(f)] for f in self.etx1.file_names], dtype=object).T
        self.energy_grid = np.vstack(self.energy_grid) 
        self.energy_pdf = np.vstack(self.energy_pdf) 
        self.dlne = np.vstack(self.dlne) 
        if self.input.sp01.track_dump:
            self.sp01 = dHybridRspecies(1, self)
        self.iter = iters(self)
        self.time = times(self)
class dHybridRgroup(SimulationGroup):
    def __init__(self, path, **sim_kwds):
        SimulationGroup.__init__(self, path, simtype=dHybridR, **sim_kwds)


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.fields import ScalarField, VectorField

from kbasic.parsing import File, Folder
from scipy.io import FortranFile
import numpy as np
from numpy import pi
from collections.abc import Iterable

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==

def parse_config_value(val:str):
    """take a string representing the value of a configuration parameter and figure out what python type it should be. 

    Args:
        val (str): The string representing the value after the = for each line of the configuration file

    Returns:
        str | int | float | np.ndarray: the python object corresponding to the configuration value
    """
    if "," in val: return np.array([float(x) for x in val.split(',')])
    elif "." in val: return float(val)
    elif val[0] in "1234567890": return int(val)
    else: return val

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class dHybridRconfig(File):
    def __init__(self, parent, mode: str|None = None) -> None:
        """Container for the information needed to initialize a dHybridR simulation

        Args:
            parent (dHybridR): the simulation this config file belongs to
            mode (str | None, optional): Represents what kind of simulation this is. Defaults to None.
        """
        self.params = []
        self.parent = parent 
        path = f"{self.parent.path}/config"
        File.__init__(self, path) #initialize as a File object
        #if the file already exists, read its contents
        if self.exists: self.read()
        #else user must supply the information
        else: self.set_interactive()

    def __setattr__(self, name, value) -> None:
        if name!="params": self.params.append(name)
        return super().__setattr__(name, value)

    def read(self) -> None:
        """method to read in the parameters in the config file and assign those to this object."""
        self.params = []
        with open(self.path, 'r') as file:
            self.lines = file.readlines()
            #for each line in the file
            for line in self.lines:
                line = line.strip()
                #skip comments and blank lines
                if any([
                    len(line)==0,
                    line.startswith("#"),
                    line.startswith("!")
                ]): continue
                # split the variable name and value and asign to this object
                name, value = (x.strip() for x in line.split("="))
                self.params.append(name)
                setattr(self, name, parse_config_value(value))
    
    def write(self) -> None:
        """write each of the parameters in the param attribute to the config file. WARNING: Overwrites config file.
        """
        #construct one line for each parameter in params
        self.lines = [f"{n}={getattr(self,n)}" for n in self.params]
        #write the file
        with open(self.path, 'w') as file: file.write("\n".join(self.lines))

    def set_interactive(self) -> None:
        print("this simulation hasn't been configured yet.")
        print("send an empty parameter to complete and write this object to file.")
        self.mode = input("What mode? ").lower()
        while len(new_param:=input("param: ").strip())>0:
            self.__setattr__(new_param, parse_config_value(input("value: ")))
            print("\n")
        self.write()

class dHybridRSnapshot:
    def __init__(
        self, 
        parent,
        i: int,
        caching: bool = False,
        verbose: bool = False
    ):
        i = i if i<len(parent.B) else len(parent.B)-1
        kwargs = {'caching':caching, 'verbose':verbose, 'parent':parent}
        print(parent, i, parent.B.x[i])
        self.B = VectorField(parent.B.x[i], parent.B.y[i], parent.B.z[i], name=parent.B.name, latex=parent.B.latex, **kwargs)
        self.u = VectorField(parent.u.x[i], parent.u.y[i], parent.u.z[i], name=parent.u.name, latex=parent.u.latex, **kwargs)
        self.E = VectorField(parent.E.x[i], parent.E.y[i], parent.E.z[i], name=parent.E.name, latex=parent.E.latex, **kwargs)
        self.density = ScalarField(parent.density[i], name=parent.density.name, latex=parent.density.latex, **kwargs)
        self.energy_grid = parent.energy_grid[i]
        self.energy_pdf = parent.energy_pdf[i]
        self.dlne = parent.dlne[i]
        self.T = sum(self.energy_grid*self.energy_pdf*self.dlne)
        self.tau = parent.tau[i]
        self.time = parent.time[i]

class dHybridRinitializer:
    def __init__(
        self,
        simulation
    ):
        self.simulation = simulation
        self.input = self.simulation.input
        self.dims = len(self.input.ncells)
        self.L = self.input.boxsize
        #parse grid size and shape from input
        self.build()

    def build(self):
        match self.dims:
            case 1:
                self.dx: float = self.L[0] / self.input.ncells[0]
                self.Nx: int = int(self.input.ncells[0])
                self.shape: tuple = self.Nx,
            case 2:
                [self.dy, self.dx] = np.array(self.L) / np.array(self.input.ncells)
                [self.Ny, self.Nx] = self.input.ncells
                self.shape = (self.Ny, self.Nx)
            case 3:
                [self.dy, self.dx, self.dz] = np.array(self.L) / np.array(self.input.ncells)
                [self.Ny, self.Nx, self.Nz] = self.input.ncells
                self.shape = (self.Ny, self.Nx, self.Nz)

    def build_B_field(self): 
        self.B = np.array([np.zeros(self.input.ncells) for i in range(2)])
    def build_u_field(self):
        self.u = np.array([np.zeros(self.input.ncells) for i in range(2)])
    def save_init_field(self, field: np.ndarray, path: str): 
        FortranFile(path, 'w').write_record(field)
    def prepare_simulation(self):
        self.build_B_field()
        self.save_init_field(self.B.T, self.simulation.path+"/input/Bfld_init.unf")
        self.build_u_field()
        self.save_init_field(self.u.T, self.simulation.path+"/input/vfld_init.unf")

class FlareWaveInit(dHybridRinitializer):
    def __init__(
            self,
            input_file,
            B0 = 1,
            Bg = 2,
            w0 = 2,
            psi0 = 0.5,
            vth = 0.1
    ):
        dHybridRinitializer.__init__(self, input_file)
        self.B0 = B0
        self.Bg = Bg 
        self.w0 = w0 
        self.psi0 = psi0

    def build_B_field(self, unknown_variable=69.12):
        x = np.arange(0, self.Nx) * self.dx
        y = np.arange(0, self.Ny) * self.dy
        Bx = np.array([
            self.B0 * (np.tanh((y - 0.25*self.L[1])/self.w0) - np.tanh((y - 0.75*self.L[1])/self.w0) - 1)
        for i in range(len(x))]).T
        By = np.array([
            (unknown_variable / self.L[0]) * np.cos(2*np.pi*x / self.L[0]) * np.sin(2*np.pi*x / self.L[0])**10
        for i in range(len(y))])
        Bz = np.sqrt(self.B0**2 + self.Bg**2 - Bx**2)
        self.B = np.array([Bx.T, By.T, Bz.T], dtype=np.float32)   


import pysim as sim
from glob import glob
import os
import shutil
import numpy as np

amp = 1.,
k = 1., np.pi

if __name__ == '__main__':
    # clean files
    for file in glob('input/*.unf'):
        os.remove(file)
    if os.path.exists("Output"): shutil.rmtree("Output")
    if os.path.exists("Restart"): shutil.rmtree("Restart")
    # make init files
    Init = sim.TurbInit(amp, k)
    Init.prepare_simulation()
    os.system("sh submit_anvil.sh")


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class Species:
    def __init__(
            self,
            name: str,
            m: float = 1.,
            q: float = 1.
        ):
        self.name = name
        self.m = m 
        self.q = q 
        self.mtq = m/q # mass to charge ratio
        self.qtm = q/m # charge to mass ratio 
        
    def __repr__(self): return self.name
class Particle:
    def __init__(
            self,
            species: Species,
            id: int
    ):
        self.species = species 
        self.id = id
        
    def __repr__(self): return f"PARTICLE {self.id} ({self.species})"


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.environment import dHybridRtemplate

from kbasic.parsing import File
import numpy as np
from datetime import datetime, date

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# fortran parsing
def input_match(input_code: str):
    """
    matches objects for dHybridR input files to equivilent python objects
    :param input_code:
    :return: python_code
    """
    match input_code:
        ###### Booleans ######
        case '.true.':
            return True
        case '.false.':
            return False
        ###### numbers ######
        case number if number.isnumeric():
            # int
            return int(number)
        case number if all([
            x.isnumeric() if len(x) > 0 else True for x in number.split('.')
        ]):
            # float
            return float(number)
        case number if all([
            "d" in number,
            "." in number,
            "/" not in number
        ]):
            # scientific notation
            v, exp = number.split("d")
            return float(v) * 10. ** int(exp)
        ###### strings ######
        case string if all([
            string[0] == '"',
            string[-1] == '"',
            "," not in string
        ]):
            return str(string[1:-1])
        ###### else ######
        case other:
            raise ValueError(f"I couldn't recognize the code you gave me\n{fortran_code}")
def python_match_input(python_code) -> str:
    """
    matches python objects to dHybridR input file equivilents
    :param python_code: python object
    :return: input equivilent
    """
    match python_code:
        case True:
            return '.true.'
        case False:
            return '.false.'
        case number if type(number) == int:
            return f"{python_code}"
        case number if type(number) in [float, np.float64, np.float32]:
            if str(number).split('.')[-1] == '0':
                return f"{int(python_code)}."
            elif 'e' in str(number):
                return str(number).replace('e', 'd')
            return f"{python_code}"
        case string if type(string) == str:
            return f'"{python_code}"'
def python2input(code):
    """
    take an iterable and convert each object to dHybridR input file equivalents
    :param code: iterable containing python objects
    :return: list of input file equivalents
    """
    if not type(code) in [tuple, list, np.ndarray]:
        return python_match_input(code)
    input_code = [f"{python_match_input(x)}," for x in code]
    return "".join(input_code)
def input2python(code):
    """
    converts dHybridR input file code into equivilent python code
    :param code:
    :return:
    """
    python_code = [input_match(x.strip(" ")) for x in code.split(',') if len(x.strip(" ")) != 0]
    if len(python_code) == 1:
        return python_code[0]
    return python_code
def is_input_header_boarder(line: str) -> bool:
    return len(chars:=np.unique([x for x in line]))==3 and all(chars == np.unique([x for x in "! -"]))
def is_input_section_header(line: str) -> bool:
    return line.startswith("!---") or line.startswith("! ---")
def is_input_species_section_header(line: str, species: int = None) -> bool:
    if species: return is_input_section_header(line) and f"for species {species}" in line 
    else: return any([is_input_section_header(line) and f"for species {sp}" in line for sp in range(10)])

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class InputParameter:
    def __init__(self, name: str, value, comment: str = None) -> None:
        self.name = name
        self.input_name = name if type(value) not in [list, tuple, np.ndarray] else f"{name}(1:{len(value)})"
        self.value = value 
        self.comment = "" if comment is None else "!"+comment
    def __str__(self) -> str:
        #            two tabs           parameter=value          make it 40 characters  then add comment
        show_string = " "*8 + f"{self.input_name}={python2input(self.value)}".ljust(40)+self.comment
        return show_string if self.value is not None else f"!{show_string}"
    def __repr__(self) -> str: return f"{self.name}: {self.value}"

class InputSection:
    def __init__(self, lines: list) -> None:
        self.lines = lines
        self.section_header: str = self.lines[0]
        #check if its an empty section
        while len(self.lines)>1 and self.lines[1].startswith("!"): del self.lines[1]
        if len(self.lines)==1: return None
        #else parse the name and make sure its bound correctly
        self.section_name: str = self.lines[1].strip()
        assert self.lines[2].strip()=="{", f"Please check this input file, improper format detected in section {self.section_name}"
        assert self.lines[-1].strip()=="}", f"Please check this input file, improper format detected in section {self.section_name}"
        self.params = {}
        for line in self.lines[3:-1]:
            #check for comments
            if line.strip().startswith("!"): continue
            #otherwise its code
            [code, *comment] = line.split("!") if "!" in line else [line.strip(), None]
            #pls don't put multiple !s on a code line :(
            comment = comment[0]
            name, value = code.strip().split("=")
            #if its a list itll have the dumb lil (1:lenth) thing
            if "(" in name: name = name.split("(")[0]
            new_param = InputParameter(name, input2python(value), comment = None if comment is None else comment.strip())
            self.params[name] = new_param
            setattr(self, name, new_param)
    def __repr__(self) -> str: return "\n".join(self.lines)
    def __str__(self) -> str: return "\n".join([
        self.section_header,
        self.section_name,
        "{",
        "\n".join([str(p) for p in self.params.values()]),
        "}"
    ])

class SpeciesInput:
    def __init__(self, lines: list, i: int) -> None:
        self.lines = lines
        self.number = i
        self.sections = {}
        i = 0
        while i+1<len(self.lines):
            line = self.lines[i]
            if is_input_section_header(line):
                sec_start = i 
                while not i+1==len(self.lines) and not is_input_section_header(self.lines[i+1]): i+=1 
                sec_end = i+1
                new_sec = InputSection(self.lines[sec_start:sec_end])
                self.sections[new_sec.section_name] = new_sec
            i+=1

        all_params = []
        for sec_name, sec in self.sections.items():
            for par_name, par in sec.params.items():
                all_params.append(par_name)
                setattr(self, par_name, par.value)

    def __str__(self) -> str: return "\n".join([str(s) for s in self.sections.values()])
    def __repr__(self) -> str: return self.name

    def save_changes(self):
        for sec in self.sections.values():
            for par_name in sec.params.keys():
                sec.params[par_name].value = getattr(self, par_name)

class dHybridRinput(File):
    header = "\n".join([
        "! -------------------------------------------------------------------------------",
        "!   dHybrid input file v 2.2",
        "!   Created for Keyan Gootkin's pysim module",
        "! -------------------------------------------------------------------------------"
    ])
    def __init__(self, path: str) -> None:
        if type(path)==list and type(path[0]==str): self.lines = path
        else:
            #init file properties
            File.__init__(self, path, master=dHybridRtemplate.path+"/input/input", executable=False)
            #if doesn't exist, make a copy from master
            # if not self.exists: self.update()
            #read in the current file
            with open(self.path, 'r') as file: self.lines = file.read().split("\n")
        #get rid of empty lines
        self.lines = [l for l in self.lines if len(l.strip())>0]
        #read the remainder to set input file properties
        self.read()
    
    def __repr__(self): return "\n".join(self.lines)
    
    def read(self) -> None:
        assert is_input_header_boarder(self.lines[0]), "must start input file with header"
        self.sections = {}
        self.species = {}
        # go through input file header
        i=0
        while not is_input_header_boarder(self.lines[i+1]): i+=1
        i+=2
        # parse sections and species
        while i+1<len(self.lines):
            line = self.lines[i]
            #check if the start of a species description
            if is_input_species_section_header(line):
                #only accepting up to 9 species as of now
                sp_num = int(line.split("for species ")[-1][0])
                sp_start = i
                #species description ends when theres a section header that is not a 
                #species section header for this species
                while not all([
                    is_input_section_header(self.lines[i+1]), 
                    not is_input_species_section_header(self.lines[i+1], species=sp_num)
                ]): i+=1 
                sp_end = i+1
                #self.sp01 is the species we just read
                setattr(
                    self, 
                    f"sp{str(sp_num).zfill(2)}", 
                    (current_sp:=SpeciesInput(self.lines[sp_start:sp_end], sp_num))
                )
                self.species[sp_num] = current_sp 
            #otherwise its the start of a normal section
            elif is_input_section_header(line):
                sec_start = i 
                while not i+1==len(self.lines) and not is_input_section_header(self.lines[i+1]): i+=1 
                sec_end = i+1
                new_sec = InputSection(self.lines[sec_start:sec_end])
                #if the section is all comments then ignore it
                if len(new_sec.lines)>1: self.sections[new_sec.section_name] = new_sec
            i+=1
        #set the section parameters to be self.parameter 
        #(all repeat parameters should be in species objects)
        all_params = []
        for sec_name, sec in self.sections.items():
            for par_name, par in sec.params.items():
                all_params.append(par_name)
                setattr(self, par_name, par.value)
        assert len(np.unique(all_params))==len(all_params), "you have a duplicate parameter in your input file! >:-("
            
    def save_changes(self) -> None:
        #set section parameters from self.parameters so that user can set the input parameters
        #to change the input section parameters and save those changes
        for sec in self.sections.values():
            for par_name in sec.params.keys():
                sec.params[par_name].value = getattr(self, par_name)
        for sp in self.species.values(): sp.save_changes()
        #seperate out sections to put before the species sections
        sections_not_all_species = [sec for sec in self.sections.values() if sec.section_name!='diag_species_total']
        #set lines
        head = self.header.split("\n")
        first_sections = "\n".join([str(sec) for sec in sections_not_all_species]).split("\n")
        species_sections  = "\n".join([str(sp) for sp in self.species.values()]).split("\n")
        final_section = str(self.sections['diag_species_total']).split("\n") 
        self.lines = np.r_[
            head,
            first_sections,
            species_sections,
            final_section      
        ]
        #write lines
        with open(self.path, 'w') as file: file.write("\n".join(self.lines))

class dHybridRout(File):
    def __init__(self, path: str) -> None:
        super().__init__(path)
        if self.exists: 
            self.read()
            self.dt: datetime = (self.end - self.start)
            self.runtime: float = self.dt.seconds / 60. / 60.

    def read(self):
        File.read(self)
        self.start = datetime.now()
        self.end = datetime.now()
        for i,li in enumerate(self.lines):
            line: list[str] = li.strip().split()
            match line:
                case ['Run', 'started', *_]:
                    [_, _, start_day, _, month, _, year, _, hour, _] = line
                    current_day = int(start_day)
                    [minute, _, second, _, milisecond] = self.lines[i+1].strip().split()
                    self.start = datetime(int(year), int(month), int(start_day), int(hour), int(minute), int(second))
                    current = self.start
                case ['ITER', *_]:
                    previous = current 
                    [*_, hour, _, minute, _, second, _, milisecond] = line
                    if int(hour)<previous.hour: current_day += 1
                    current = datetime(int(year), int(month), current_day, int(hour), int(minute), int(second))
                    self.end = current
                case ['Run', 'terminated', 'successfully', *_]: 
                    [_, _, _, day, _, month, _, year, _] = line
                    [hour, _, minute, _, second, _, milisecond] = self.lines[i+1].strip().split()
                    self.end = datetime(int(year), int(month), int(day), int(hour), int(minute), int(second))


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.environment import dHybridRtemplate
from kbasic.parsing import File

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                           Definitions                           <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
main_submit_script: str = """

module load intel
module load hdf5

if [ -z ${SLURM_JOB_ID+x} ]; then
    n1=`awk '/node_number/{split($1, a, "="); split(a[2], b, ","); print b[1]}' input/input`
    n2=`awk '/node_number/{split($1, a, "="); split(a[2], b, ","); print b[2]}' input/input`
    n3=`awk '/node_number/{split($1, a, "="); split(a[2], b, ","); print b[3]}' input/input`
    n1=${n1:=1}
    n2=${n2:=1}
    n3=${n3:=1}
    ntasks=`echo $n1*$n2*$n3 | bc`
    CDN=${PWD##*/}
    taskspernode=128
    nodes=`echo $ntasks/$taskspernode | bc`
    remainder=`echo $ntasks%$taskspernode | bc`
    if (( remainder > 0 )); then
        nodes=`echo $nodes+1 | bc`
    fi
    echo x_tasks=$n1, y_tasks=$n2, z_tasks=$n3, n_tasks=$ntasks, nodes=$nodes
    sbatch --ntasks=$ntasks --nodes=$nodes --job-name=$CDN --output=$CDN.out --error=$CDN.err $0

else
    echo $SLURM_JOB_ID
    mpirun dHybridR > out
fi
"""

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class AnvilSubmitScript(File):
    def __init__(
        self,
        path: str,
        queue: str = "wholenode",
        time_limit: str = "24:00:00",
        email: str|None = None,
        allocation: str = "ast180008"
    ) -> None:
        File.__init__(self, path, master=dHybridRtemplate.path+"/submit_anvil.sh")
        assert queue.lower() in ["wholenode", "debug"], f"Queue: {queue} not available, please choose either wholenode or debug"
        self.queue = queue.lower()
        self.time_limit = time_limit
        self.email = email
        self.allocation = allocation
        self.build()

    def __str__(self) -> str: return self.text
    def __repr__(self) -> str: return self.text

    def build(self) -> None: 
        self.header = "\n".join([
            "#!/bin/bash",
            f"#SBATCH -p {self.queue}      # Queue (partition) name",
            f"#SBATCH -t {self.time_limit}        # Run time (hh:mm:ss)",
            f"{'#' if self.email else '##'}SBATCH --mail-user={self.email}",
            f"{'#' if self.email else '##'}SBATCH --mail-type=all    # Send email at begin and end of job",
            f"#SBATCH -A {self.allocation}       # Allocation name (req'd if you have more than 1)"
        ])
        self.text = self.header + main_submit_script

    def write(self) -> None:
        with open(self.path, 'w') as file: file.write(self.text)


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.environment import simulationDir

from kbasic.parsing import Folder, File
from kbasic.user_input import yesno
from os.path import isdir
from matplotlib.pyplot import cm as cmaps

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class GenericSimulation:
    def __init__(
            self, 
            path:str, 
            template:str|Folder=None,
            caching:bool=False,
            verbose:bool=True,
        ) -> None:
        """A generic class for all simulations

        Args:
            path (str): the location of the simulation
            template (str | Folder, optional): the location of the template this simulation is based on. Defaults to None.
            caching (bool, optional): whether or not to set up a cache to store data in. Defaults to False.
            verbose (bool, optional): whether to print a bunch of bullshit. Defaults to True.
        """
        self.template = template
        self.verbose = verbose
        #setup cache
        if self.verbose: print("caching is ON..." if caching else "caching is OFF...")
        self.caching = caching 
        self.cache: dict = {}
        #make sure the simulation exists
        if verbose: print(f"Finding path: {path}")
        self.path: str = path
        self.dir = Folder(path)
        self.name = self.dir.name
        #if the given path doesn't exist, check the default simulation directory 
        if not self.dir.exists: 
            if self.verbose: print(f"No simulation found in {path}, checking default simulation directory: {simulationDir.path}...")
            default_path: str = f"{simulationDir.path}/{self.name}"
            self.path = default_path
            self.dir = Folder(self.path)
            #if simulation isn't in default simulation directory either copy a template to that location or raise an error
            if not self.dir.exists:
                if yesno(f"No such simulation exists, would you like to copy \ntemplate: {template.name}, \nto location: {self.path}?\n"):
                    self.create()
                else: raise FileNotFoundError("Please create simulation and try again")
    
    def create(self):
        self.template.copy(self.path)

class SimulationGroup(Folder):
    def __init__(self, path: str, simtype = GenericSimulation, **sim_kwds) -> None: 
        Folder.__init__(self, path)
        self.simulations = {x.split('/')[-1]:simtype(x, **sim_kwds) for x in self.children if isdir(x) and File(x+"/input/input").exists}
    
    def __repr__(self) -> str: return self.name+'\n'+'-'*20+"\n"+"\n".join([f"{k}: {repr(v)}" for k, v in self.simulations.items()])
    def __len__(self) -> int: return len(self.simulations)
    def __getitem__(self, item): return self.simulations[item]
    def __iter__(self):
        self.index = 0
        return self
    def __next__(self):
        if self.index < len(self):
            i = self.index
            self.index += 1
            return list(self.values())[i]
        else: raise StopIteration

    def sort_by(self, key: str) -> None:
        new_simulations = {v.__dict__[key]: v for k, v in sorted(self.simulations.items(), key = lambda item: item[1].__dict__[key])}
        if len(new_simulations) != len(self.simulations): raise KeyError(f"the simulation value: {key} is not unique in {self.name}, please provide a unique key to sort by.")
        self.simulations = new_simulations

    def colorer(self, cmap=cmaps.plasma) -> list: return [cmap(i / (len(self)+.1)) for i in range(len(self))]
    def labeler(self) -> list[str]: return [x.name for x in self.simulations.values()]
    def plotter(self, cmap=cmaps.plasma) -> list[tuple]: return [(l_i, c_i, sim_i) for l_i, c_i, sim_i in zip(self.labeler(), self.colorer(cmap=cmap), self.simulations.values())]

    def items(self): return self.simulations.items()
    def values(self): return self.simulations.values()
    def keys(self): return self.simulations.keys()


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
#pysim imports
from kgsim.tristan.tristan import Tristan, TristanParticleSpecies, TristanParticleQuantity
from kgsim.fields import VectorField
#non-pysim imports
import numpy as np

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class MaskedTristanParticleQuantity(TristanParticleQuantity):
    def __init__(
            self,
            param_key,
            mask_param,
            mask_value,
            species = None,
            parent = None,
            caching = False,
            verbose: bool = False,
            name: str = None, 
            latex: str = None
    ):
        TristanParticleQuantity.__init__(
            self,
            param_key,
            parent = parent,
            caching = caching,
            verbose = verbose,
            name = name, 
            latex = latex 
        )
        self.mask_param = mask_param
        self.mask_value = mask_value
        self.species = species

    def __getitem__(self, item):
        mask = np.where(getattr(self.species, self.mask_param)[item]==self.mask_value)
        return super().__getitem__(item)[mask]

class SubPopulation(TristanParticleSpecies):
    def __init__(
            self,
            parent_species,
            param,
            value
    ):
        self.parent_species = parent_species
        self.param = param 
        self.value = value
        TristanParticleSpecies.__init__(
            self, 
            self.parent_species.name, 
            parent = self.parent_species.parent, 
            m = self.parent_species.m, 
            q = self.parent_species.q, 
            N = self.parent_species.N, 
            sig = self.parent_species.sig
        )
        kwargs = {'caching':self.parent.caching, 'verbose':self.parent.verbose, 'parent':self.parent}
        self.x = MaskedTristanParticleQuantity(
            'x'+self.sig, self.param, self.value, 
            species=self.parent_species, name="x", latex="$x$", **kwargs
        )
        self.y = MaskedTristanParticleQuantity(
            'y'+self.sig, self.param, self.value, 
            species=self.parent_species, name="y", latex="$y$", **kwargs
        )
        self.z = MaskedTristanParticleQuantity(
            'z'+self.sig, self.param, self.value, 
            species=self.parent_species, name="z", latex="$z$", **kwargs
        )
        self.v = VectorField(
            MaskedTristanParticleQuantity(
                "u"+self.sig, self.param, self.value, 
                species=self.parent_species, name="vx", latex="$v_x$", **kwargs
            ),
            MaskedTristanParticleQuantity(
                "v"+self.sig, self.param, self.value, 
                species=self.parent_species, name="vy", latex="$v_y$", **kwargs
            ),
            MaskedTristanParticleQuantity(
                "w"+self.sig, self.param, self.value, 
                species=self.parent_species, name="vz", latex="$v_z$", **kwargs
            ),
            name="v", latex=r"$\vec{v}$"
        )
        self.gamma = MaskedTristanParticleQuantity(
            "gamma"+self.sig, self.param, self.value, 
            species=self.parent_species, name="gamma", latex=rf"$\gamma_{self.sig}$", **kwargs
        )
        self.ch = MaskedTristanParticleQuantity(
            "ch"+self.sig, self.param, self.value, 
            species=self.parent_species, name="ch", latex=rf"weight$", **kwargs
        )

class KappaSim(Tristan):
    def __init__(
        self,
        path: str,
        caching: bool = False,
        verbose: bool = False,
        compressed: bool = False
    ):
        Tristan.__init__(
            self,
            path,
            caching=caching,
            verbose=verbose,
            compressed=compressed
        )
        self.kappa = self.input.kappa.value 
        self.kappa_density = self.input.kappadens.value
        self.ions.kappa = SubPopulation(self.ions, "ch", self.kappa_density)
        self.electrons.kappa = SubPopulation(self.electrons, "ch", self.kappa_density)
        self.thermal_density = max(np.unique(self.ions.ch[0]))
        self.ions.thermal = SubPopulation(self.ions, "ch", self.thermal_density)
        self.electrons.thermal = SubPopulation(self.electrons, "ch", self.thermal_density)


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kgsim.fields import ScalarField, VectorField
from kgsim.simulation import GenericSimulation
from kgsim.dhybridr.io import InputParameter

from kbasic.parsing import Folder
from glob import glob
import numpy as np
from h5py import File

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                           Definitions                           <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# keys are parameters and values are the type of file to read those from
tristan_param_file_convert = {
    'bx': 'flds.tot', 
    'by': 'flds.tot', 
    'bz': 'flds.tot', 
    'dens': 'flds.tot', 
    'densi': 'flds.tot', 
    'ex': 'flds.tot', 
    'ey': 'flds.tot', 
    'ez': 'flds.tot', 
    'jx': 'flds.tot', 
    'jy': 'flds.tot', 
    'jz': 'flds.tot', 
    'v3x': 'flds.tot', 
    'v3xi': 'flds.tot', 
    'v3y': 'flds.tot', 
    'v3yi': 'flds.tot', 
    'v3z': 'flds.tot', 
    'v3zi': 'flds.tot',
    'che': 'prtl.tot', 
    'chi': 'prtl.tot', 
    'gammae': 'prtl.tot', 
    'gammai': 'prtl.tot', 
    'inde': 'prtl.tot', 
    'indi': 'prtl.tot', 
    'proce': 'prtl.tot', 
    'proci': 'prtl.tot', 
    'time': 'prtl.tot', 
    'ue': 'prtl.tot', 
    'ui': 'prtl.tot', 
    've': 'prtl.tot', 
    'vi': 'prtl.tot', 
    'we': 'prtl.tot', 
    'wi': 'prtl.tot', 
    'xe': 'prtl.tot', 
    'xi': 'prtl.tot', 
    'ye': 'prtl.tot', 
    'yi': 'prtl.tot', 
    'ze': 'prtl.tot', 
    'zi': 'prtl.tot',
    'acool': 'param', 
    'bphi': 'param', 
    'btheta': 'param', 
    'c': 'param', 
    'c_omp': 'param', 
    'caseinit': 'param', 
    'cooling': 'param', 
    'delgam': 'param', 
    'dlapion': 'param', 
    'dlaplec': 'param', 
    'dummy': 'param', 
    'gamma0': 'param', 
    'interval': 'param', 
    'istep': 'param', 
    'istep1': 'param', 
    'me': 'param', 
    'mi': 'param', 
    'mx': 'param', 
    'mx0': 'param', 
    'my': 'param', 
    'my0': 'param', 
    'mz0': 'param', 
    'ntimes': 'param', 
    'pltstart': 'param', 
    'ppc0': 'param', 
    'qi': 'param', 
    'sigma': 'param', 
    'sizex': 'param', 
    'sizey': 'param', 
    'stride': 'param', 
    'testendion': 'param', 
    'testendlec': 'param', 
    'teststarti': 'param', 
    'teststartl': 'param', 
    'time': 'param', 
    'torqint': 'param', 
    'walloc': 'param', 
    'xinject2': 'param',
    'gamma': 'spectl', 
    'gmax': 'spectl', 
    'gmin': 'spectl', 
    'spece': 'spectl', 
    'specerest': 'spectl', 
    'specp': 'spectl', 
    'specprest': 'spectl', 
    'xsl': 'spectl'
}

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
def tristan_loader(
        sim_path: str, inds: list[int], params: list[str], 
        zfill_level: int = 3, padding: int = 3
) -> dict:
    # Colby wrote this one I think?
    output = {}
    param_files = [tristan_param_file_convert[pi] for pi in params]
    # for each file we need to read
    for pf in np.unique(param_files):
        for i in inds:
            with File(f"{sim_path}/output/{pf}.{str(i+1).zfill(zfill_level)}") as file:
                # for each parameter requested
                for pi in params:
                    if pi not in file: continue # only take the params in the active file
                    P = file[pi][:].copy()
                    if P.ndim==3: P = P[0, padding : -padding + 1, padding : -padding + 1]
                    output[pi] = np.append(output[pi], P, axis=0) if pi in output else P
    return output

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class TristanInput:
    def __init__(
            self,
            path: str,
    ):
        """A class for reading/storing info from Tristan input files.

        Args:
            path (str): path to the input file.
        """
        self.path = path
        with open(self.path, 'r') as file:
            self.lines = file.readlines()
        for line in self.lines:
            line = line.strip()
            if line.startswith(("#", '<')) or len(line)==0: continue
            param, *valcom = line.split('=')
            valcom = "=".join(valcom)
            value, comment = valcom.split('#') if '#' in valcom else (valcom.strip(), None)
            param, value = param.strip(), value.strip()
            setattr(self, param, InputParameter(param, float(value), comment=comment))

class TristanScalarField(ScalarField):
    def __init__(
            self,
            param_key: str,
            parent = None,
            caching: bool = False,
            verbose: bool = False,
            name: str = None,
            latex: str = None,
            zfill_level: int = 3,
            padding: int = 3
    ):
        """A ScalarField object specialized for Tristan output files.

        Args:
            param_key (str): key used to store info in output files
            parent (Tristan, optional): The parent simultion from which to read outputs. Defaults to None.
            caching (bool, optional): Whether to store field values after reading. Defaults to False.
            verbose (bool, optional): Doesn't do much rn but might be good for debugging later. Defaults to False.
            name (str, optional): Name of field to be used in e.g. saving plots of. Defaults to None.
            latex (str, optional): Name of field to be used when using latex. Defaults to None.
            zfill_level (int, optional): Number of 0's to use when constructing file names. Defaults to 3.
            padding (int, optional): number of overflow cells surrounding simulation domain. Defaults to 3.
        """
        ScalarField.__init__(
            self,
            None,
            parent=parent,
            caching=caching,
            verbose=verbose,
            name=name,
            latex=latex
        )
        self.param_key = param_key
        self.zfill_level = zfill_level
        self.padding = padding
        self.single = False
        self.file_prefix = tristan_param_file_convert[self.param_key]
        self.file_names = glob(f"{self.parent.path}/output/{tristan_param_file_convert[self.param_key]}.*")

    def reader(self, fname: str, item: int):
        output = tristan_loader(self.parent.path, [item], [self.param_key], zfill_level=self.zfill_level, padding=self.padding)
        return output[self.param_key]

class TristanParticleQuantity(TristanScalarField):
    def __init__(            
            self,
            param_key: str,
            parent = None,
            caching: bool = False,
            verbose: bool = False,
            name: str = None,
            latex: str = None
    ):
        """Object for storing Tristan particle data

        Args:
            param_key (str): key used to access output data
            parent (Tristan, optional): Simulation from which output is drawn. Defaults to None.
            caching (bool, optional): Whether to store data after reading. Defaults to False.
            verbose (bool, optional): Doesn't do much for now. Defaults to False.
            name (str, optional): Name to use in e.g. filenames. Defaults to None.
            latex (str, optional): name to use in latex. Defaults to None.
        """
        TristanScalarField.__init__(
            self,
            param_key,
            parent = parent,
            caching = caching,
            verbose = verbose,
            name = name,
            latex = latex,
            padding = 0
        )

class TristanParticleSpecies:
    def __init__(
            self,
            name: str,
            parent = None,
            m: float = 1.,
            q: float = 1.,
            N: int = None,
            sig: str = ""
    ):
        """Object to store data for a particle species from a Tristan simulation.

        Args:
            name (str): Name of the particle species
            parent (Tristan, optional): Simulation from which to draw data. Defaults to None.
            m (float, optional): mass. Defaults to 1..
            q (float, optional): charge. Defaults to 1..
            N (int, optional): number of particles. Defaults to None.
            sig (str, optional): signature to add to the end of keys e.g. 'i' for ions. Defaults to "".
        """
        self.name = name 
        self.m = m 
        self.q = q 
        self.N = N
        self.sig = sig
        self.parent = parent
        if parent is not None:
            # load actual parameters
            kwargs = {'caching':parent.caching, 'verbose':parent.verbose, 'parent':parent}
            self.x = TristanParticleQuantity("x"+sig, name="x", latex="$x$", **kwargs)
            self.y = TristanParticleQuantity("y"+sig, name="y", latex="$y$", **kwargs)
            self.z = TristanParticleQuantity("z"+sig, name="z", latex="$z$", **kwargs)
            self.v = VectorField(
                TristanParticleQuantity("u"+sig, name="vx", latex="$v_x$", **kwargs),
                TristanParticleQuantity("v"+sig, name="vy", latex="$v_y$", **kwargs),
                TristanParticleQuantity("w"+sig, name="vz", latex="$v_z$", **kwargs),
                name="v", latex=r"$\vec{v}$"
            )
            self.gamma = TristanParticleQuantity("gamma"+sig, name="gamma", latex=rf"$\gamma_{sig}$", **kwargs)
            self.ch = TristanParticleQuantity("ch"+sig, **kwargs)
            self.Bx = TristanParticleQuantity("bx", name="Bx", latex="$B_x$", **kwargs)

            # self.density,_,_ = np.array([np.histogram2d(self.x[i], self.y[i], bins=[self.parent.Nx, self.parent.Ny], weights=self.ch) for i in range(len(self.x))])

class Tristan(GenericSimulation):
    def __init__(
            self,
            path: str,
            template: str|Folder = None,
            caching: bool = False,
            verbose: bool = False,
            compressed: bool = False
    ):
        """Object to interact with Tristan simulations.

        Args:
            path (str): path to simulation top-level
            template (str | Folder, optional): template from which to compare. Defaults to None.
            caching (bool, optional): whether to store data after reading it in. Defaults to False.
            verbose (bool, optional): Doesn't do much rn. Defaults to False.
            compressed (bool, optional): Whether this is a full simulation or just some timestamps. Defaults to False.
        """
        self.compressed = compressed
        GenericSimulation.__init__(self, path, template=template, caching=caching, verbose=verbose)
        self.parse_input()
        self.parse_output()

    def __len__(self): return len(glob(f"{self.path}/output/flds.tot.*"))
    def __repr__(self): return f"Tristan Simulation: {self.path}"

    def parse_input(self):
        """Read input file and construct relevant quantities
        """
        self.input = TristanInput(self.path+"/input")
        # time
        self.c = self.input.c.value
        self.steps = np.arange(0, self.input.last.value, self.input.interval.value)
        w_pe = self.steps[-1] * self.c / self.input.c_omp.value
        om_ce = w_pe * np.sqrt(self.input.sigma.value)
        mratio = self.input.mi.value / self.input.me.value
        om_ci = om_ce / mratio
        self.time = np.round(self.steps / om_ci, 0)
        # grids
        self.Nx = int(self.input.mx0.value)
        self.Ny = int(self.input.my0.value)
        self.Nz = int(self.input.mz0.value)
        self.shape = (self.Nx, self.Ny, self.Nz)
        d_xe = self.Nx * self.input.istep.value / self.input.c_omp.value
        d_xi = d_xe / self.input.mi.value 
        d_ye = self.Ny * self.input.istep.value / self.input.c_omp.value
        d_yi = d_ye / self.input.mi.value
        self.x = np.linspace(0, d_xe, self.Nx)
        self.y = np.linspace(0, d_ye, self.Ny)
        self.dx = np.diff(self.x)[0]
        self.dy = np.diff(self.y)[0]

    def parse_output(self):
        """Read output files and construct relevant data structures.
        """
        kwargs = {'caching':self.caching, 'verbose':self.verbose, 'parent':self}
        self.B = VectorField(
            TristanScalarField("bx", name='bx', latex='$B_x$', **kwargs),
            TristanScalarField("by", name='by', latex='$B_y$', **kwargs),
            TristanScalarField("bz", name='bz', latex='$B_z$', **kwargs),
            name='magnetic', latex=r"$\vec{B}$", **kwargs
        )
        self.E = VectorField(
            TristanScalarField("ex", name='ex', latex='$E_x$', **kwargs),
            TristanScalarField("ey", name='ey', latex='$E_y$', **kwargs),
            TristanScalarField("ez", name='ez', latex='$E_z$', **kwargs),
            name='electric', latex=r"$\vec{E}$", **kwargs
        )
        self.J = VectorField(
            TristanScalarField("jx", name='jx', latex='$J_x$', **kwargs),
            TristanScalarField("jy", name='jy', latex='$J_y$', **kwargs),
            TristanScalarField("jz", name='jz', latex='$J_z$', **kwargs),
            name='electric', latex=r"$\vec{J}$", **kwargs
        )
        self.u = VectorField(
            TristanScalarField("v3x", name='ux', latex='$u_x$', **kwargs),
            TristanScalarField("v3y", name='uy', latex='$u_y$', **kwargs),
            TristanScalarField("v3z", name='uz', latex='$u_z$', **kwargs),
            name='bulk-flow', latex=r"$\vec{u}$", **kwargs
        )
        self.density = TristanScalarField("dens", name='density', latex=r"$\rho$", **kwargs)
        self.ions = TristanParticleSpecies("proton", parent=self, m=self.input.mi.value, q=1, sig='i')
        self.electrons = TristanParticleSpecies("electron", parent=self, m=self.input.me.value, q=-1, sig='e')

    def load(self, params: list[str], item: int|slice = slice(None, None)):
        """loads data from params at item times.

        Args:
            params (list[str]): keys representing data to load
            item (int | slice, optional): which timestamps to load data for. Defaults to slice(None, None).

        Returns:
            data (dict): a dictionary with params as keys and values corresponding to the data at each time in item.
        """
        match item:
            case int(): 
                items = [item]
            case slice(): 
                items = [
                    x for x in range(
                        0 if item.start is None else item.start, 
                        len(self) if item.stop is None else item.stop,
                        1 if item.step is None else item.step
                        )
                    ]
        return tristan_loader(self.path, items, params)


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Imports                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
from kplot import show, show_video
from kbasic.bar import verbose_bar
from kbasic.parsing import Folder, File
from kbasic.typing import Number, Iterable
from glob import glob
import numpy as np
from h5py import File as h5File
from functools import cached_property
from os.path import isdir, isfile
from matplotlib.pyplot import gca
from matplotlib.axes import Axes
from matplotlib.cm import plasma as default_cmap


# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                            Functions                            <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
def calc_psi(Bx, By, dx, dy):
    psi = np.zeros(Bx.shape)
    psi[1:,0] = np.cumsum(Bx[1:,0])*dy
    psi[:,1:] = (psi[:,0] - np.cumsum(By[:,1:], axis=1).T*dx).T
    return psi

def ddx(F,dx): return np.gradient(F, dx)[-1]
def ddy(F,dy): return np.gradient(F, dy)[-2] 
def ddz(F,dz): return np.gradient(F, dz)[-3]


def intdx(F,dx): return np.cumsum(F, axis=1) * dx
def intdy(F,dy): return np.cumsum(F, axis=0) * dy
def Az(Bx, By, dx=1, dy=1):
    Az = np.zeros(Bx.shape)
    Az[1:] = intdy(Bx[1:], dy)
    Az[:,1:] = (Az[:,0]-intdx(By[:,1:], dx).T).T
    return Az

# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
# >-|===|>                             Classes                             <|===|-<
# !==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==!==
class ScalarField:
    """
    a special class of arrays used to efficiently interact with the fields output by simulations
    ________
    ~Inputs~
    * source - str | array-like
        the source of the scalar field. Can be a file, a folder full of files, or an array like object.
    ___________
    ~Atributes~
    * single - bool
        whether or this is a single field as opposed to a collection of fields
    * shape - tuple[int]
        the shape of output arrays a la numpy arrays
    * ndims - int 
        the number of dimensions in the output arrays

    ===FILE MODE=== 
    * path - str
        path containing field files 
    * file_names - list[str]
        the files containing fields
    * caching - bool
        whether or not to store file outputs for later use, more memory intensive but fewer file accesses
    * cache - dict
        a dictionary to store file outputs for later use
    * reader - function
        the function used to read files

    ===ARRAY MODE===
    * array - numpy.ndarray
        the array representing the field
    """
    def __init__(
        self, 
        source: str|Folder|File|list|np.ndarray, 
        parent = None,
        caching: bool = False,
        verbose: bool = False,
        name: str = None, 
        latex: str = None
    ) -> None:
        self.name: str = name 
        self.latex: str = latex
        self.verbose: bool = verbose
        self.parent = parent
        # if parent:
        #     self.di = tuple(np.array(parent.input.boxsize, dtype=float) / np.array(parent.input.ncells, dtype=int))
        #setup cache
        self.caching: bool = caching
        self.cache: dict = {}
        #find the correct constructor
        match source:
            #if its a folder
            case Folder():
                self.single = False
                example_file = File(source.children[0])
                if example_file.extension=="h5": self._from_folder_of_h5(source.path)
            case str() if isdir(source): 
                self.single = False
                files = glob(source+"/*")
                extension = files[0].split(".")[-1]
                if extension=="h5": self._from_folder_of_h5(source)
            #otherwise its a file
            case File():
                self.single = True
                if source.extension=="h5": self._from_h5(source.path)
            case str() if isfile(source): 
                self.single = True
                extension = source.split(".")[-1] 
                if extension=="h5": self._from_h5(source)
                else: self._from_csv(source)
            #or if its already been read
            case np.ndarray(): 
                self.single = True
                self._from_numpy(source)
            case list(): 
                self.single = True
                self._from_numpy(np.array(source))
    def __len__(self) -> int: return 1 if self.single else len(self.file_names)
    def __iter__(self):
        assert not self.single, "Cannot iterate through single scalar field"
        self.index = 0
        return self
    def __next__(self):
        if self.index < len(self):
            i = self.index
            self.index += 1
            return self[i]
        else: raise StopIteration
    def __getitem__(self, item: int|slice|tuple|list) -> np.ndarray:
        if self.single: return self.array[item]
        match item:
            case int(): 
                return self.cache[item] if self.caching and item in self.cache.keys() else self.reader(self.file_names[item], item)
            case slice():
                item_iters = [
                    i for i in range(
                        item.start if not item.start is None else 0, 
                        item.stop if not item.stop is None else len(self), 
                        item.step if not item.step is None else 1
                    )
                ]
                return np.array([
                    self.cache[i] if self.caching and i in self.cache.keys() else self.reader(self.file_names[i], i) for i in item_iters
                ])
            case tuple()|list(): return np.array([
                self.cache[i] if self.caching and i in self.cache.keys() else self.reader(self.file_names[i], i) for i in item
            ])

    def _from_folder_of_h5(self, path:str) -> None: 
        self.path = path.path if isinstance(path, Folder) else path
        self.file_names: list = sorted(glob(path + "/*.h5"))
        self.reader: function = self._read_h5_file
        self.shape = self[0].shape
        self.ndims = len(self.shape)
        self.size = np.prod(self.shape) * len(self)
    def _from_h5(self, file:str) -> None:
        self.file = file.path if isinstance(file,File) else file
        self.array = self._read_h5_file(file, 0)
        self.shape = self.array.shape 
        self.ndims = len(self.shape)
        self.size = np.prod(self.shape) * len(self)
    def _read_h5_file(self, file:str, item) -> np.ndarray:
        with h5File(file, 'r') as f:
            output = np.array(f["DATA"][:])
            #GODDMANIT I HATE THAT IT DOES Y,X and not X,Y
            if self.caching: self.cache[item] = output
            return output
    
    def _from_csv(self) -> None:
        self.single = True
    def _read_csv_file(self) -> None: pass
    
    def _from_numpy(self, array:np.ndarray) -> None:
        self.single = True
        self.array = array
        self.shape = array.shape
        self.ndims = len(self.shape)

    def show(self, item:int, **kwargs) -> None: show(self[item],**kwargs)
    
    def movie(self, norm='none', cmap=default_cmap, alter_func=None,**kwrg) -> None:
        @show_video(name=self.name, latex=self.latex, norm=norm, cmap=cmap)
        def reveal_thyself(s,alter_func=alter_func, **kwargs): 
            return np.array([self[i] for i in range(len(self))]) if alter_func is None else np.array([alter_func(self[i]) for i in range(len(self))])
        reveal_thyself(self if self.parent is None else self.parent, alter_func=alter_func,**kwrg)

class VectorField:
    def __init__(
            self, 
            *components, 
            caching: bool =False, 
            verbose: bool =False,
            name: str = None, 
            latex: str = None,
            parent = None,
            parallel: str = 'z'
        ) -> None:
        latex = "".join([c for c in latex if c not in r"$\{}"])
        self.name = name
        self.latex = latex 
        self.parent = parent
        # if parent: 
        #     self.dx, self.dy = parent.dx, parent.dy
        self.verbose = verbose 
        self.caching = caching
        child_kwargs = {'parent':parent, 'verbose':verbose, 'caching':caching}
        if len(components)==1 and type(path:=components[0])==str:
            components = (
                ScalarField(path+"/x", name=name+"_x_component", latex=f"${latex}_x$", **child_kwargs), 
                ScalarField(path+"/y", name=name+"_y_component", latex=f"${latex}_y$", **child_kwargs), 
                ScalarField(path+"/z", name=name+"_z_component", latex=f"${latex}_z$", **child_kwargs)
            )
        self.ndims = len(components)
        assert 1<self.ndims<4, f"Only 2-3 components are supported. {len(components)} were given"
        component_names = "xyz"
        self.components = []
        for name,val in zip(component_names, components): 
            comp = ScalarField(val, caching=self.caching) if type(val)==str else val
            self.components.append(comp)
            setattr(self, name, comp)
        self.size = sum([c.size for c in self.components])
        self.set_parallel(parallel)

    def __len__(self) -> int: return min([len(self.x), len(self.y), len(self.z)])
    def __abs__(self) -> np.ndarray:
        homo = all([len(self.x[i])==len(self.x[0]) for i in range(len(self))])
        return np.array([
            np.sqrt(sum([
                c[i]**2 for c in self.components
            ])) for i in verbose_bar(range(len(self)), self.verbose, desc="taking magnitude...")
        ], dtype=float if homo else object)
    def __getitem__(self, item: int|slice) -> np.ndarray:
        match type(item):
            case int(): return np.array([c[item] for c in self.components])
            case slice():
                item_iters = [
                    i for i in range(
                        item.start if not item.start is None else 0, 
                        item.stop if not item.stop is None else len(self), 
                        item.step if not item.step is None else 1
                    )
                ]
                return np.array([
                    [
                        c[i] for c in self.components
                    ] for i in item_iters
                ])
    
    def dot(self, other) -> np.ndarray: 
        if isinstance(other, VectorField):
            assert self.ndims==3, "only 3D vector fields can be dotted at this time"
            assert other.ndims==3, "Can only dot into a 3D vector field at this time"
            if self.verbose: print("constructing A...")
            A = np.array([
                [
                    [
                        [self.x[k][i,j], self.y[k][i,j], self.z[k][i,j]]
                    for i in range(self.shape[0])]
                for j in range(self.shape[1])] 
            for k in verbose_bar(range(len(self)), self.verbose, desc="constructing A...")])
            B = np.array([
                [
                    [
                        [other.x[k][i,j], other.y[k][i,j], other.z[k][i,j]]
                    for i in range(other.shape[0])]
                for j in range(other.shape[1])] 
            for k in verbose_bar(range(len(other)), self.verbose, desc="constructing B...")])
            return np.sum(A * B, axis=2)
    def cross(self, other, k:int) -> np.ndarray:
        if type(other)==VectorField: 
            assert self.ndims==3, "only 3D vector fields can be crossed at this time"
            assert other.ndims==3, "Can only cross into a 3D vector field at this time"
            return np.array([
                self.y[k]*other.z[k] - self.z[k]*other.y[k], 
                self.z[k]*other.x[k] - self.x[k]*other.z[k],
                self.x[k]*other.y[k] - self.y[k]*other.x[k]
            ])
    @cached_property
    def potential(self):
        match len(self.x.shape):
            case 2: return np.array([Az(self.x[i], self.y[i], dx=self.dx, dy=self.dy) for i in range(len(self))])
    def curlz(self, item: int|slice, order: int = 2):
        match type(item):
            case int(): return curlz(self.x[item], self.y[item], order=order)
            case slice():
                item_iters = [
                    i for i in range(
                        item.start if not item.start is None else 0, 
                        item.stop if not item.stop is None else len(self), 
                        item.step if not item.step is None else 1
                    )
                ]
                return np.array([curlz(self.x[i], self.y[i], order=order) for i in item_iters])
    def div(self, item: int|slice, order: int = 2):
        if not item: return np.array([divergence(self.x[i], self.y[i], dx=self.dx, dy=self.dy, order=order) for i in range(len(self))])
        match type(item):
            case int(): return divergence(self.x[item], self.y[item], dx=self.dx, dy=self.dy, order=order)
            case slice():
                item_iters = [
                    i for i in range(
                        item.start if not item.start is None else 0, 
                        item.stop if not item.stop is None else len(self), 
                        item.step if not item.step is None else 1
                    )
                ]
                return np.array([divergence(self.x[i], self.y[i], dx=self.dx, dy=self.dy, order=order) for i in item_iters])
                

    def calc_Jz(self, item=None, verbose=True):
        if not item: self.Jz = np.array([np.nanstd(j) for j in verbose_bar(self.curlz(slice(0,len(self))), verbose)])
        elif type(item) in [int, slice]: self.Jz = np.nanstd(self.curlz(item), axis=(1,2))
        else: raise TypeError(f"calc_Jz only takes ints, slices, or None for item, not {type(item)}-type objects")

    def calc_perp(self, item=None) -> np.ndarray: 
        if not item:
            self.perp = np.array([
                np.hypot(self.perpendicular[0][j], self.perpendicular[1][j]) 
                for j in verbose_bar(range(len(self)), self.verbose, desc="perpendicularizing")
            ])
        elif type(item)==int: 
            self.perp = np.hypot(self.perpendicular[0][item], self.perpendicular[1][item])
        elif type(item)==slice: 
            item_iters = [
                i for i in range(
                    item.start if item.start else 0,
                    item.stop if item.stop else len(self),
                    item.step if item.step else 1
                )
            ]
            self.perp = np.array([np.hypot(self.perpendicular[0][j], self.perpendicular[1][j]) for j in item_iters])
        else: raise TypeError(f"calc_perp only takes ints, slices, or None for item, not {type(item)}-type objects")
    @cached_property
    def psi(self) -> np.ndarray: return np.array([
        calc_psi(Bx, By, self.dx, self.dy) for Bx, By in zip(self.x, self.y)
    ])       
    def set_parallel(self, component:str) -> None:
        match component.lower():
            case 'x':
                self.parallel = self.x 
                self.perpendicular = self.y, self.z 
            case 'y':
                self.parallel = self.y 
                self.perpendicular = self.x, self.z 
            case 'z':
                self.parallel = self.z 
                self.perpendicular = self.x, self.y 
    def movie(self, mode='mag', norm='none', cmap=default_cmap, **kwrg) -> None:
        match mode.lower():
            case 'mag'|'magnitude'|'abs':
                @show_video(name=self.name+"_magnitude", latex=fr"$|{self.name}|$", norm=norm, cmap=cmap)
                def reveal_thyself(s, **kwargs): return abs(self)
            case 'perp'|'perpendicular':
                @show_video(name=self.name+"_perp", latex=fr"${self.name}_\perp$", norm=norm, cmap=cmap)
                def reveal_thyself(s, **kwargs): return self.perp                
            case 'par'|'parallel':
                @show_video(name=self.name+"_par", latex=fr"${self.name}_\parallel$", norm=norm, cmap=cmap)
                def reveal_thyself(s, **kwargs): return np.array([self.parallel[i] for i in range(len(self))])
        reveal_thyself(self if self.parent is None else self.parent, **kwrg)
    def quiver(
        self,
        ind: int,
        ax: Axes = None,
        #x/y axes
        density: float = 10,
        x: np.ndarray|None = None,
        y: np.ndarray|None = None,
        transpose: bool = False,
        #everything else goes into matplotlib command
        **kwargs
    ):
        if not ax: ax = gca()
        match x, y:
            case None, None:
                dx = int(self.x.shape[1] // density)
                x = np.arange(0, self.x.shape[1], dx)
                dy = int(self.y.shape[0] // density)
                y = np.arange(0, self.x.shape[0], dy)
        # plot data
        if not transpose:
            ax.quiver(x, y, self.y[ind][::dy,::dx].T, self.x[ind][::dy,::dx].T, **kwargs)
        else: 
            ax.quiver(y, x, self.x[ind][::dy,::dx], self.y[ind][::dy,::dx], **kwargs)
class NablaOperator:
    __ins = None
    def __new__(cls):
        if cls.__ins is None:
            cls.__ins = super().__new__(cls)
        return cls.__ins
    def parse_grid_di(self, Field, di=None) -> tuple:
        match di:
            case _ if type(di) in Iterable.types: return tuple(di)
            case _ if type(di) in Number.types: return (di, di, di) 
            case None:
                match Field:
                    case np.ndarray(): pass 
                    case ScalarField(parent=None): return (1, 1, 1)
                    case ScalarField(ndims=3): return (Field.parent.dz, Field.parent.dy, Field.parent.dx)
                    case ScalarField(ndims=2): return (Field.parent.dy, Field.parent.dx)
                    case VectorField(): return parse_grid_di(Field.x, di=di)
    def __mul__(self, other, di=None): # divergence
        match other:
            case VectorField():
                d = self.parse_grid_di(other.x)
                dz,dy,dx = d if len(d)==3 else tuple([np.inf, *d])
                dFxdx = ddx(other.x[:], dx)
                dFydy = ddy(other.y[:], dy)
                dFzdz = ddz(other.z[:], dz) if other.ndims==3 else 0
                return dFxdx + dFydy + dFzdz
    def x(self, other, di=1): # curl
        match other:
            case VectorField(): 
                d = self.parse_grid_di(other.x)
                dz,dy,dx = d if len(d)==3 else tuple([np.inf, *d])
                dt = 1 if not other.parent else other.parent.dt
                dFzdy = ddy(other.z[:], dy)
                dFydz = ddz(other.y[:], dz) if other.ndims==3 else 0 
                dFxdz = ddz(other.x[:], dz) if other.ndims==3 else 0 
                dFzdx = ddx(other.z[:], dx) 
                dFydx = ddx(other.y[:], dx)
                dFxdy = ddy(other.x[:], dy)
                return VectorField(
                    dFzdy - dFydz,
                    dFxdz - dFzdx,
                    dFydx - dFxdy,
                    caching=other.caching, verbose=other.verbose, parent=other.parent, 
                    name=f"Curl of {other.name}", latex=fr"$\Nabla \times {other.latex}"
                )
            case np.ndarray():
                d = self.parse_grid_di(other, di)
                dz,dy,dx = d if len(d)==3 else tuple([np.inf, *d])
                dFzdy = ddy(other[-3], dy) if len(d)==3 else 0
                dFydz = ddz(other[-2], dz) if len(d)==3 else 0 
                dFxdz = ddz(other[-1], dz) if len(d)==3 else 0 
                dFzdx = ddx(other[-3], dx) if len(d)==3 else 0
                dFydx = ddx(other[-2], dx)
                dFxdy = ddy(other[-1], dy)
                return np.array([
                    dFydx - dFxdy,
                    dFxdz - dFzdx,
                    dFzdy - dFydz
                ]) if len(d)==3 else dFydx - dFxdy
        
    def __call__(self, *args, **kwds): # gradient
        match args:
            case (ScalarField(),): 
                F: ScalarField = args[0]
                arr: np.ndarray = F[:]
                di: tuple[float] = F.di if 'di' not in kwds.keys() else kwds['di']
                gradarr: np.ndarray = np.gradient(arr, di)
                # assume that the 0th axis is time, as will be the case with all simulation outputs, but not necessarily numpy arrays
                # assume that the last axis is x
                return VectorField(
                    *gradarr[:0:-1],
                    caching=F.caching, verbose=F.verbose, name="grad-"+F.name, latex=fr"$\Nabla {F.latex[1:]}", parent=F.parent
                )
            case _: print("bad arguments : "+str(args)+str(kwds))

Nabla = NablaOperator()
Del = NablaOperator()
Grad = NablaOperator()